#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    unmanic.library.py

    Written by:               Josh.5 <jsunnex@gmail.com>
    Date:                     06 Feb 2022, (12:11 PM)

    Copyright:
           Copyright (C) Josh Sunnex - All Rights Reserved

           Permission is hereby granted, free of charge, to any person obtaining a copy
           of this software and associated documentation files (the "Software"), to deal
           in the Software without restriction, including without limitation the rights
           to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
           copies of the Software, and to permit persons to whom the Software is
           furnished to do so, subject to the following conditions:

           The above copyright notice and this permission notice shall be included in all
           copies or substantial portions of the Software.

           THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
           EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
           MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
           IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
           DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
           OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
           OR OTHER DEALINGS IN THE SOFTWARE.

"""
import random

from unmanic.config import Config
from unmanic.libs import common
from unmanic.libs.unmodels import EnabledPlugins, Libraries, LibraryPluginFlow, Plugins, Tags, Tasks


def generate_random_library_name():
    names = [
        "Willes", "Here", "Helry", "Vyncent", "Burgwy", "Homas Yournet", "Roguy Eldys", "George Ewes", "Hearda",
        "Mathye Gedde", "Wynfre", "Gauwill", "Aldhert", "Ryany", "Reward", "Atwulf", "Amer", "Alten Yourner", "Reda", "Oled",
        "Anthohn Dene", "Rarder", "Artin Borne", "Eadwean", "Freyny Loray", "Breda", "Gauwalt Nynsell", "Lodwy", "Exam",
        "Alters Corby", "Wilhye", "Gery", "Raffin", "Ceolbehrt", "Jamath", "George Sone", "Geoffrey Nette", "Eadund", "Dunne",
        "Gilda", "Aered", "Lafa", "Eadulf", "Eanmaed", "Cyni", "Draffin", "Nichye", "Reder", "Aldwid", "Conbad", "Munda",
        "Willex", "Ichohn", "Orkold", "Gyleon", "Ealard", "Helmund", "Nother", "Bertio", "Phamund Erett", "Cuthre", "Aewald",
        "Aehehrt", "Folke", "Ales", "Chury Kypwe", "Liamund", "Rewalt Wyne", "Arryn", "Charlip", "Georguy", "Lare", "Aenward",
        "Eanwald", "Ashwid", "Britheard", "Cholas", "Eolhed", "Anwulf", "Eorcorht", "Piersym", "Godre", "Edward", "Dreder",
        "Geoffry", "Wyny", "Hardwy", "Witio", "Grewis", "Chilew", "Gare", "Arnwulf", "Masym Arren", "Iged", "Uwan", "Coenwy",
        "Saefa", "Thiles", "Cyne", "Exard", "Ichas Horne", "Rewilh Morley", "Edmur Ferry", "Wine", "Ered", "Lacio", "Elres",
        "Gaenbyrtf", "Stomund", "Riffin Maley", "Thiliam Save", "Walda", "Giles Drighte", "Robern Finchey", "Wulfa", "James",
        "Stiny Fane", "Driffin", "Andrers", "Beorhtio", "Balda", "Warder", "Bealdu", "Dene", "Andren", "Stephye", "Ealcar",
        "Richye Corby", "Ament Anes", "Tharry", "Germund", "Ralphye Payney"
    ]
    adjectives = [
        "awesome", "adorable", "abounding", "aspiring", "beloved", "blue", "blissful", "creamy", "cavernous", "content",
        "droopy", "excited", "enchanted", "enormous", "extroverted", "exciting", "gullible", "gaseous", "grumpy", "giant",
        "handsome", "hefty", "harmless", "happy", "hairy", "humdrum", "invincible", "illiterate", "inexperienced", "impolite",
        "illustrious", "impartial", "innocent", "jovial", "juvenile", "joyful", "jumpy", "jagged", "joyous", "kooky", "large",
        "likeable", "mountainous", "momentous", "minty", "nocturnal", "nautical", "organic", "overcooked", "productive",
        "plush", "polished", "queasy", "quirky", "quintessential", "reminiscent", "remarkable", "ragged", "rowdy", "soggy",
        "sudden", "scandalous", "secretive", "spry", "squiggly", "smooth", "sulky", "scented", "spicy", "sticky", "slushy",
        "symptomatic", "tart", "turbulent", "tiresome", "typical", "xyloid", "xanthic", "zealous", "zany",
    ]
    return "{name}, the {adjective} library".format(name=random.choice(names), adjective=random.choice(adjectives))


class Library(object):
    """
    Library

    Contains all data pertaining to a library

    """

    def __init__(self, library_id: int):
        # Ensure library ID is not 0
        if library_id < 1:
            raise Exception("Library ID cannot be less than 1")
        self.model = Libraries.get_or_none(id=library_id)
        if not self.model:
            raise Exception("Unable to fetch library with ID {}".format(library_id))

    @staticmethod
    def get_all_libraries():
        """
        Return a list of all libraries

        :return:
        """
        # Fetch default library path from
        from unmanic.config import Config
        default_library_path = Config().get_library_path()
        if not default_library_path:
            default_library_path = common.get_default_library_path()

        # Fetch all libraries from DB
        configured_libraries = Libraries.select()

        # Ensure that at least the default path was added.
        # If the libraries path is empty, then we should add the default path
        if not configured_libraries:
            default_library = {
                'id':                 1,
                'name':               generate_random_library_name(),
                'path':               default_library_path,
                'locked':             False,
                "enable_remote_only": False,
                "enable_scanner":     False,
                "enable_inotify":     False,
                'tags':               [],
            }
            Libraries.create(**default_library)
            return [default_library]

        # Loop over results
        default_library = []
        libraries = []
        for lib in configured_libraries:
            # Always update the default library path
            if lib.id == 1 and lib.path != default_library_path:
                lib.path = default_library_path
                lib.save()
            # Create library config dictionary
            library_config = {
                'id':                 lib.id,
                'name':               lib.name,
                'path':               lib.path,
                'locked':             lib.locked,
                'enable_remote_only': lib.enable_remote_only,
                'enable_scanner':     lib.enable_scanner,
                'enable_inotify':     lib.enable_inotify,
                'tags':               [],
            }
            # Append tags
            for tag in lib.tags.order_by(Tags.name):
                library_config['tags'].append(tag.name)

            # Keep the default library separate
            if lib.id == 1:
                default_library.append(library_config)
                continue
            libraries.append(library_config)

        # Return the list of libraries sorted by name
        return default_library + sorted(libraries, key=lambda d: d['name'])

    @staticmethod
    def within_library_count_limits(frontend_messages=None):
        # Fetch level from session
        from unmanic.libs.session import Session
        s = Session()
        s.register_unmanic()
        if s.level > 1:
            return True

        # Fetch all enabled plugins
        library_count = Libraries.select().count()

        # Ensure enabled plugins are within limits
        # Function was returned above if the user was logged in and able to use infinite
        if library_count > s.library_count:
            # If the frontend messages queue was included in request, append a message
            if frontend_messages:
                frontend_messages.put(
                    {
                        'id':      'libraryEnabledLimits',
                        'type':    'error',
                        'code':    'libraryEnabledLimits',
                        'message': '',
                        'timeout': 0
                    }
                )
            return False
        # If the frontend messages queue was included in request, remove the notification as we are currently within limits
        if frontend_messages:
            frontend_messages.remove_item('libraryEnabledLimits')
        return True

    @staticmethod
    def create(data: dict):
        """
        Create a new library

        :param data:
        :return:
        """
        # Ensure ID is removed from data for a create
        if 'id' in data:
            del data['id']
        new_library = Libraries.create(**data)
        return Library(new_library.id)

    @staticmethod
    def export(library_id):
        from unmanic.libs.plugins import PluginsHandler

        # Read the library
        library_config = Library(library_id)

        # Get list of enabled plugins with their settings
        enabled_plugins = []
        for enabled_plugin in library_config.get_enabled_plugins(include_settings=True):
            enabled_plugins.append({
                'plugin_id':  enabled_plugin.get('plugin_id'),
                'has_config': enabled_plugin.get('has_config'),
                'settings':   enabled_plugin.get('settings'),
            })

        # Create plugin flow
        plugin_flow = {}

        plugin_handler = PluginsHandler()
        for plugin_type in plugin_handler.get_plugin_types_with_flows():
            plugin_flow[plugin_type] = []
            flow = plugin_handler.get_enabled_plugin_flows_for_plugin_type(plugin_type, library_id)
            for f in flow:
                plugin_flow[plugin_type].append(f.get('plugin_id'))

        return {
            "plugins":        {
                "enabled_plugins": enabled_plugins,
                "plugin_flow":     plugin_flow,
            },
            "library_config": {
                "name":               library_config.get_name(),
                "path":               library_config.get_path(),
                'enable_remote_only': library_config.get_enable_remote_only(),
                'enable_scanner':     library_config.get_enable_scanner(),
                'enable_inotify':     library_config.get_enable_inotify(),
                'tags':               library_config.get_tags(),
            },
        }

    def __remove_enabled_plugins(self):
        """
        Remove all enabled plugins

        :return:
        """
        query = EnabledPlugins.delete()
        query = query.where(EnabledPlugins.library_id == self.model.id)
        return query.execute()

    def __trim_plugin_flow(self, plugin_ids: list):
        """
        Trim the plugin flow removing entries not in the given plugin ids list

        :param plugin_ids:
        :return:
        """
        query = LibraryPluginFlow.delete()
        query = query.where((LibraryPluginFlow.library_id == self.model.id) & (LibraryPluginFlow.plugin_id.not_in(plugin_ids)))
        return query.execute()

    def __remove_associated_tasks(self):
        """
        Remove all tasks associated with a library

        :return:
        """
        Tasks.delete().where(Tasks.library_id == self.model.id).execute()

    def get_id(self):
        return self.model.id

    def get_name(self):
        return self.model.name

    def set_name(self, value):
        self.model.name = value

    def get_path(self):
        return self.model.path

    def set_path(self, value):
        self.model.path = value

    def get_locked(self):
        return self.model.locked

    def set_locked(self, value):
        self.model.locked = value

    def get_enable_remote_only(self):
        return self.model.enable_remote_only

    def set_enable_remote_only(self, value):
        self.model.enable_remote_only = value

    def get_enable_scanner(self):
        return self.model.enable_scanner

    def set_enable_scanner(self, value):
        self.model.enable_scanner = value

    def get_enable_inotify(self):
        return self.model.enable_inotify

    def set_enable_inotify(self, value):
        self.model.enable_inotify = value

    def get_priority_score(self):
        return self.model.priority_score

    def set_priority_score(self, value):
        self.model.priority_score = value

    def get_tags(self):
        return_tags = []
        for tag in self.model.tags.order_by(Tags.name):
            return_tags.append(tag.name)
        return return_tags

    def set_tags(self, value):
        # Create any missing tags
        for tag_name in value:
            # Do not update any current tags with on_conflict_replace() as this will also change their IDs
            # Instead, just ignore them
            Tags.insert(name=tag_name).on_conflict_ignore().execute()
        # Create a SELECT query for all tags with the listed names
        tags_select_query = Tags.select().where(Tags.name.in_(value))
        # Clear out the current linking table of tags linked to this library
        # Add new links for each tag that was fetched matching the provided names
        self.model.tags.add(tags_select_query, clear_existing=True)

    def get_enabled_plugins(self, include_settings=False):
        """
        Get all enabled plugins for this library

        :return:
        """
        # Fetch enabled plugins for this library
        query = self.model.enabled_plugins.select(Plugins, EnabledPlugins.library_id)
        query = query.join(Plugins, join_type='LEFT OUTER JOIN', on=(EnabledPlugins.plugin_id == Plugins.id))
        query = query.order_by(Plugins.name)

        from unmanic.libs.unplugins import PluginExecutor
        plugin_executor = PluginExecutor()

        # Extract required data
        enabled_plugins = []
        for enabled_plugin in query.dicts():
            # Check if plugin is able to be configured
            has_config = False
            plugin_settings, plugin_settings_meta = plugin_executor.get_plugin_settings(enabled_plugin.get('plugin_id'),
                                                                                        library_id=self.model.id)
            if plugin_settings:
                has_config = True
            # Add plugin to list of enabled plugins
            item = {
                'plugin_id':   enabled_plugin.get('plugin_id'),
                'name':        enabled_plugin.get('name'),
                'description': enabled_plugin.get('description'),
                'icon':        enabled_plugin.get('icon'),
                'has_config':  has_config,
            }
            if include_settings:
                item['settings'] = plugin_settings
            enabled_plugins.append(item)

        return enabled_plugins

    def get_plugin_flow(self):
        """
        Fetch the plugin flow for a library

        :return:
        """
        plugin_flow = {}
        from unmanic.libs.plugins import PluginsHandler
        plugin_handler = PluginsHandler()
        from unmanic.libs.unplugins import PluginExecutor
        plugin_ex = PluginExecutor()
        for plugin_type in plugin_ex.get_all_plugin_types():
            # Ignore types without flows
            if not plugin_type.get('has_flow'):
                continue

            # Create list of plugins in this plugin type
            plugin_flow[plugin_type.get('id')] = []
            plugin_modules = plugin_handler.get_enabled_plugin_modules_by_type(plugin_type.get('id'), library_id=self.model.id)
            for plugin_module in plugin_modules:
                plugin_flow[plugin_type.get('id')].append(
                    {
                        "plugin_id":   plugin_module.get("plugin_id"),
                        "name":        plugin_module.get("name", ""),
                        "author":      plugin_module.get("author", ""),
                        "description": plugin_module.get("description", ""),
                        "version":     plugin_module.get("version", ""),
                        "icon":        plugin_module.get("icon", ""),
                    }
                )

        return plugin_flow

    def __set_default_plugin_flow_priority(self, plugin_list):
        from unmanic.libs.unplugins import PluginExecutor
        plugin_executor = PluginExecutor()
        from unmanic.libs.plugins import PluginsHandler
        plugin_handler = PluginsHandler()

        # Fetch current items
        configured_plugin_ids = []
        query = LibraryPluginFlow.select().where(LibraryPluginFlow.library_id == self.model.id)
        for flow_item in query:
            configured_plugin_ids.append(flow_item.plugin_id.plugin_id)

        for plugin in plugin_list:
            # Ignore already configured plugins
            if plugin.get('plugin_id') in configured_plugin_ids:
                continue
            plugin_info = plugin_handler.get_plugin_info(plugin.get('plugin_id'))
            plugin_priorities = plugin_info.get('priorities')
            if plugin_priorities:
                # Fetch the plugin info back from the DB
                plugin_info = Plugins.select().where(Plugins.plugin_id == plugin.get("plugin_id")).first()
                # Fetch all plugin types in this plugin
                plugin_types_in_plugin = plugin_executor.get_all_plugin_types_in_plugin(plugin.get("plugin_id"))
                # Loop over the plugin types in this plugin
                for plugin_type in plugin_types_in_plugin:
                    # get the plugin runner function name for this runner
                    plugin_type_meta = plugin_executor.get_plugin_type_meta(plugin_type)
                    runner_string = plugin_type_meta.plugin_runner()
                    if plugin_priorities.get(runner_string) and int(plugin_priorities.get(runner_string, 0)) > 0:
                        # If the runner has a priority set and that value is greater than 0 (default that wont set anything),
                        # Save the priority
                        PluginsHandler.set_plugin_flow_position_for_single_plugin(
                            plugin_info,
                            plugin_type,
                            self.model.id,
                            plugin_priorities.get(runner_string)
                        )

    def set_enabled_plugins(self, plugin_list: list):
        """
        Update the list of enabled plugins

        :param plugin_list:
        :return:
        """
        # Remove all enabled plugins
        self.__remove_enabled_plugins()

        # Add new repos
        data = []
        plugin_ids = []
        for plugin_info in plugin_list:
            plugin = Plugins.get(plugin_id=plugin_info.get('plugin_id'))
            plugin_ids.append(plugin.id)
            if plugin:
                data.append({
                    "library_id":  self.model.id,
                    "plugin_id":   plugin,
                    "plugin_name": plugin.name,
                })

        # Delete all plugin flows for plugins not to be enabled for this library
        self.__trim_plugin_flow(plugin_ids)

        # Insert plugins
        EnabledPlugins.insert_many(data).execute()

        # Add default flow for newly added plugins
        self.__set_default_plugin_flow_priority(plugin_list)

    def save(self):
        """
        Save the data for this library

        :return:
        """
        # Save changes made to model
        save_result = self.model.save()

        # If this is the default library path, save to config.library_path object also
        if self.get_id() == 1:
            config = Config()
            config.set_config_item('library_path', self.get_path())

        return save_result

    def delete(self):
        """
        Delete the current library

        :return:
        """
        # Ensure we can never delete library ID 1 (the default library)
        if self.get_id() == 1:
            raise Exception("Unable to remove the default library")

        # Ensure we are not trying to delete a locked library
        if self.get_locked():
            raise Exception("Unable to remove a locked library")

        # Remove all enabled plugins
        self.__remove_enabled_plugins()

        # Remove all plugin flows
        self.__trim_plugin_flow([])

        # Delete all tasks with matching library_id
        self.__remove_associated_tasks()

        # Remove the library entry
        return self.model.delete_instance(recursive=True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    unmanic.session.py

    Written by:               Josh.5 <jsunnex@gmail.com>
    Date:                     10 Mar 2021, (5:20 PM)

    Copyright:
           Copyright (C) Josh Sunnex - All Rights Reserved

           Permission is hereby granted, free of charge, to any person obtaining a copy
           of this software and associated documentation files (the "Software"), to deal
           in the Software without restriction, including without limitation the rights
           to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
           copies of the Software, and to permit persons to whom the Software is
           furnished to do so, subject to the following conditions:

           The above copyright notice and this permission notice shall be included in all
           copies or substantial portions of the Software.

           THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
           EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
           MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
           IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
           DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
           OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
           OR OTHER DEALINGS IN THE SOFTWARE.

"""
import base64
import datetime
import pickle
import random
import time

import requests

from unmanic import config
from unmanic.libs import common, unlogger
from unmanic.libs.singleton import SingletonType
from unmanic.libs.unmodels import Installation


class RemoteApiException(Exception):
    """
    RemoteApiException
    Custom exception for errors contacting the remote Unmanic API
    """

    def __init__(self, message, status_code):
        super().__init__(f"Session Error - Remote API [CODE: {status_code}]: {message}")


class Session(object, metaclass=SingletonType):
    """
    Session

    Manages the Unmanic applications session for unlocking
    features and fetching data from the Unmanic site API.

    """

    """
    level - The user auth level
    Set level to 0 by default
    """
    level = 0

    """
    non supporter library count
    """
    library_count = 2

    """
    non supporter linked installations count
    """
    link_count = 5

    """
    picture_uri - The user avatar
    """
    picture_uri = ''

    """
    name - The user's name
    """
    name = ''

    """
    email - The user's email
    """
    email = ''

    """
    created - The timestamp when the session was created
    """
    created = None

    """
    last_check - The timestamp when the session was last checked
    """
    last_check = None

    """
    uuid - This installation's UUID
    """
    uuid = None

    """
    user_access_token - The access token to authenticate requests with the Unmanic API
    """
    user_access_token = None

    """
    user_refresh_token - The refresh token for acquiring an updated access token
    """
    user_refresh_token = None

    """
    session_cookies - A stored copy of the session cookies to persist between restarts
    """
    session_cookies = None

    def __init__(self, *args, **kwargs):
        unmanic_logging = unlogger.UnmanicLogger.__call__()
        self.logger = unmanic_logging.get_logger(__class__.__name__)
        self.timeout = 30
        self.dev_api = kwargs.get('dev_api', None)
        self.requests_session = requests.Session()
        self.logger.info('Initialising new session object')

    def __created_older_than_x_days(self, days=1):
        # (86400 = 24 hours)
        seconds = (days * 86400)
        # Get session expiration time
        time_now = time.time()
        time_when_session_expires = self.created + seconds
        # Check that the time create is less than 24 hours old
        if time_now < time_when_session_expires:
            return False
        return True

    def __check_session_valid(self):
        """
        Ensure that the session is valid.
        A session is only valid for a limited amount of time.
        After a session expires, it should be re-acquired.

        :return:
        """
        # Last checked less than a min ago... just keep the current session.
        # This check is only to prevent spamming requests when the site API is unreachable
        # Only check in every 40 mins (2400s) minimum. Never ignore a checkin for more than 45 mins (2700s)
        ### if self.last_check and 2700 > (time.time() - self.last_check) < 2400:
        if self.last_check and 45 > (time.time() - self.last_check) < 40:
            return True
        # If the session has never been created, return false
        if not self.created:
            return False
        # Check if the time the session was created is less than 1 day old
        if not self.__created_older_than_x_days(days=2):
            # Only try to recreate the session once a day
            return True
        self.logger.debug('Session no longer valid')
        return False

    def __update_created_timestamp(self):
        """
        Update the session "created" timestamp.

        :return:
        """
        # Get the time now in seconds
        seconds = time.time()
        # Create a seconds offset of some random number between 300 (5 mins) and 900 (15 mins)
        seconds_offset = random.randint(300, 900 - 1)
        # Set the created flag with the seconds variable plus a random offset to avoid people joining
        #   together to register if the site goes down
        self.created = (seconds + seconds_offset)
        # Print only the accurate update time in debug log
        created = datetime.datetime.fromtimestamp(seconds)
        self.logger.debug('Updated session at %s', str(created))

    def __fetch_installation_data(self):
        """
        Fetch installation data from DB

        :return:
        """
        # Fetch installation
        db_installation = Installation()
        try:
            # Fetch a single row (get() will raise DoesNotExist exception if no results are found)
            current_installation = db_installation.select().order_by(Installation.id.asc()).limit(1).get()
        except Exception as e:
            # Create settings (defaults will be applied)
            self.logger.debug('Unmanic session does not yet exist... Creating.')
            db_installation.delete().execute()
            current_installation = db_installation.create()

        self.uuid = str(current_installation.uuid)
        self.level = int(current_installation.level)
        self.picture_uri = str(current_installation.picture_uri)
        self.name = str(current_installation.name)
        self.email = str(current_installation.email)
        self.created = current_installation.created if current_installation.created else None
        if isinstance(self.created, datetime.datetime):
            self.created = self.created.timestamp()

        self.__update_session_auth(access_token=current_installation.user_access_token,
                                   session_cookies=current_installation.session_cookies)

    def __store_installation_data(self, force_save_access_token=False):
        """
        Store installation data in DB to persist reboot

        :return:
        """
        if self.uuid:
            db_installation = Installation.get_or_none(uuid=self.uuid)
            db_installation.level = self.level
            db_installation.picture_uri = self.picture_uri
            db_installation.name = self.name
            db_installation.email = self.email
            db_installation.created = self.created
            if self.user_access_token or force_save_access_token:
                db_installation.user_access_token = self.user_access_token
            if self.user_refresh_token or force_save_access_token:
                db_installation.user_refresh_token = self.user_refresh_token
            if self.session_cookies or force_save_access_token:
                db_installation.session_cookies = self.session_cookies
            db_installation.save()

    def __reset_session_installation_data(self):
        """
        Reset stored session data

        :return:
        """
        self.logger.debug('Resetting session installation data.')
        self.level = 0
        self.picture_uri = ''
        self.name = ''
        self.email = ''
        self.created = time.time()
        self.user_access_token = None
        self.user_refresh_token = None
        self.__store_installation_data(force_save_access_token=True)
        self.__clear_session_auth()

    def __update_session_auth(self, access_token=None, session_cookies=None):
        # Update session headers
        if access_token:
            self.user_access_token = access_token
            self.requests_session.headers.update({'Authorization': self.user_access_token})
        # Update session cookies
        if session_cookies:
            self.session_cookies = session_cookies
            try:
                self.requests_session.cookies = pickle.loads(base64.b64decode(session_cookies))
            except Exception as e:
                self.logger.error('Error trying to reload session cookies - %s', str(e))

    def __clear_session_auth(self):
        self.requests_session.cookies.clear()
        self.requests_session.headers.update({'Authorization': ''})

    def get_installation_uuid(self):
        """
        Returns the installation UUID as a string.
        If it does not yet exist, it will create one.

        :return:
        """
        if not self.uuid:
            self.__fetch_installation_data()
        return self.uuid

    def get_supporter_level(self):
        """
        Returns the supporter level

        :return:
        """
        if not self.level:
            self.__fetch_installation_data()
        return self.level

    def get_site_url(self):
        """
        Set the Unmanic application site URL
        :return:
        """
        api_proto = "https"
        api_domain = "api.unmanic.app"
        if self.dev_api:
            return self.dev_api
        return "{0}://{1}".format(api_proto, api_domain)

    def set_full_api_url(self, api_prefix, api_version, api_path):
        """
        Set the API path URL

        :param api_prefix:
        :param api_version:
        :param api_path:
        :return:
        """
        api_versioned_path = "{}/v{}".format(api_prefix, api_version)
        return "{0}/{1}/{2}".format(self.get_site_url(), api_versioned_path, api_path)

    def api_get(self, api_prefix, api_version, api_path):
        """
        Generate and execute a GET API call.

        :param api_prefix:
        :param api_version:
        :param api_path:
        :return:
        """
        u = self.set_full_api_url(api_prefix, api_version, api_path)
        r = self.requests_session.get(u, timeout=self.timeout)
        if r.status_code > 403:
            # There is an issue with the remote API
            raise RemoteApiException(f"GET request failed for {u}", r.status_code)
        if r.status_code == 401:
            # Verify the token. Refresh as required
            self.logger.debug("Auto exec token verification (api_get)")
            token_verified = self.verify_token()
            # If successful, then retry request
            if token_verified:
                r = self.requests_session.get(u, timeout=self.timeout)
                if r.status_code > 403:
                    # There is an issue with the remote API
                    raise RemoteApiException(f"GET request still failed for {u}", r.status_code)
            else:
                self.logger.debug('Failed to verify auth (api_get)')
        return r.json(), r.status_code

    def api_post(self, api_prefix, api_version, api_path, data):
        """
        Generate and execute a POST API call.

        :param api_prefix:
        :param api_version:
        :param api_path:
        :param data:
        :return:
        """
        u = self.set_full_api_url(api_prefix, api_version, api_path)
        r = self.requests_session.post(u, json=data, timeout=self.timeout)
        if r.status_code > 403:
            # There is an issue with the remote API
            raise RemoteApiException(f"POST request failed for {u}", r.status_code)
        if r.status_code == 401:
            # Verify the token. Refresh as required
            self.logger.debug("Auto exec token verification (api_post)")
            token_verified = self.verify_token()
            # If successful, then retry request
            if token_verified:
                r = self.requests_session.post(u, json=data, timeout=self.timeout)
                if r.status_code > 403:
                    # There is an issue with the remote API
                    raise RemoteApiException(f"POST request still failed for {u}", r.status_code)
            else:
                self.logger.debug('Failed to verify auth (api_post)')
        return r.json(), r.status_code

    def verify_token(self):
        if not self.user_access_token:
            # No valid tokens exist
            return False
        # Check if access token is valid
        u = self.set_full_api_url('support-auth-api', 1, 'user_auth/verify_token')
        r = self.requests_session.get(u, timeout=self.timeout)
        if r.status_code in [202]:
            # Token is valid
            return True
        elif r.status_code > 403:
            # Issue with server... Just carry on with current access token can't fix that here.
            raise RemoteApiException(f"Token verification request failed for {u}", r.status_code)

        # Access token is not valid. Refresh it.
        self.logger.debug('Unable to verify authentication token. Refreshing...')
        d = {"refreshToken": self.user_refresh_token}
        u = self.set_full_api_url('support-auth-api', 1, 'user_auth/token')
        r = self.requests_session.post(u, json=d, timeout=self.timeout)
        if r.status_code in [202]:
            # Token refreshed
            # Store the updated access token
            response = r.json()
            self.__update_session_auth(access_token=response.get('data', {}).get('accessToken'))
            # Store the updated refresh token
            self.user_refresh_token = response.get('data', {}).get('refreshToken')
            # Store the updated session cookies
            self.session_cookies = base64.b64encode(pickle.dumps(self.requests_session.cookies)).decode('utf-8')
            self.__store_installation_data()
            return True
        elif r.status_code > 403:
            # Issue was with server... Just carry on with current access token can't fix that here.
            raise RemoteApiException(f"Token refresh request failed for {u}", r.status_code)
        elif r.status_code in [403]:
            # Log this failure in the debug logs
            self.logger.info('Failed to refresh access token.')
            response = r.json()
            for message in response.get('messages', []):
                self.logger.info('Remote Message: %s', message)
        # Just blank the class attribute.
        # It is fine for requests to be sent with further requests.
        # User will appear to be logged out.
        self.user_access_token = None
        return False

    def fetch_user_data(self):
        # Set default level to 0
        updated_level = 0
        response, status_code = self.api_get('support-auth-api', 1, 'user_auth/user_info')
        if status_code > 403:
            # Failed to fetch data from server. Ignore this for now. Will try again later.
            raise RemoteApiException("Failed to fetch user info from user_auth/user_info", status_code)
        if status_code in [200, 201, 202] and response.get('success'):
            # Get user data from response data
            user_data = response.get('data', {}).get('user')
            if user_data:
                # Set name from user data
                self.name = user_data.get("name", "Valued Supporter")
                # Set avatar from user data
                self.picture_uri = user_data.get("picture_uri", "/assets/global/img/avatar/avatar_placeholder.png")
                # Set email from user data
                self.email = user_data.get("email", "")
                # Update level from response data (default back to 0)
                updated_level = int(user_data.get("supporter_level", 0))
        return updated_level

    def auth_user_account(self, force_checkin=False):
        # Don't bother if the user has never logged in
        if not self.user_access_token and not force_checkin:
            self.logger.debug('The user access token is not set add we are not being forced to refresh for one.')
            return False
        # Start by verifying the token
        token_verified = self.verify_token()
        if not token_verified and force_checkin:
            # Try to fetch token if this was the initial login
            post_data = {"uuid": self.get_installation_uuid()}
            response, status_code = self.api_post('support-auth-api', 1, 'app_auth/retrieve_app_token', post_data)
            if status_code in [200, 201, 202] and response.get('success'):
                # Store the updated access token
                self.__update_session_auth(access_token=response.get('data', {}).get('accessToken'))
                # Store the updated refresh token
                self.user_refresh_token = response.get('data', {}).get('refreshToken')
                token_verified = self.verify_token()
            elif status_code > 403:
                raise RemoteApiException("Failed to fetch initial app token frp, app_auth/retrieve_app_token", status_code)
            else:
                self.logger.info('Failed to check in with Unmanic support auth API.')
                for message in response.get('messages', []):
                    self.logger.info('Remote Message: %s', message)
        # Finally, fetch user info if the token was successfully verified
        if token_verified:
            self.level = self.fetch_user_data()
            return True
        else:
            # Set default level to 0
            self.level = 0
        return False

    def auth_trial_account(self):
        # Check if access token is valid
        d = {"uuid": self.get_installation_uuid()}
        u = self.set_full_api_url('support-auth-api', 1, 'user_auth/trial_token')
        r = self.requests_session.post(u, json=d, timeout=self.timeout)
        if r.status_code in [202]:
            # Token refreshed
            # Store the updated access token
            response = r.json()
            self.__update_session_auth(access_token=response.get('data', {}).get('accessToken'))
            # Fetch user data
            self.level = self.fetch_user_data()
            # Store the updated session cookies
            self.__store_installation_data()
            return True
        elif r.status_code > 403:
            # Issue with server... Just carry on with current access token can't fix that here.
            raise RemoteApiException(f"Trial token verification request failed for {u}", r.status_code)

    def register_unmanic(self, force=False):
        """
        Register Unmanic with site.
        This sends information about the system that Unmanic is running on.
        It also sends a unique ID.

        Based on the return information, this will set the session level.

        Return success status.

        :param force:
        :return:
        """
        # First check if the current session is still valid
        if not force and self.__check_session_valid():
            return True

        # Set now as the last time this was run (before it was actually run
        self.last_check = time.time()

        # Update the session
        settings = config.Config()
        # Fetch the installation data prior to running a session update
        self.__fetch_installation_data()

        try:
            # Build post data
            from unmanic.libs.system import System
            system = System()
            system_info = system.info()
            platform_info = system_info.get("platform", None)
            if platform_info:
                platform_info = " * ".join(platform_info)
            post_data = {
                "uuid":           self.get_installation_uuid(),
                "version":        settings.read_version(),
                "python_version": system_info.get("python", ''),
                "system":         {
                    "platform": platform_info,
                    "devices":  system_info.get("devices", {}),
                }
            }

            # Refresh user auth
            result = self.auth_user_account(force_checkin=force)
            if not result:
                # Fetch trial token
                result = self.auth_trial_account()

            # Register Unmanic
            registration_response, status_code = self.api_post('unmanic-api', 1, 'installation_auth/register', post_data)

            # Save data
            if status_code in [200, 201, 202] and registration_response.get("success"):
                self.__update_created_timestamp()
                # Persist session in DB
                self.__store_installation_data()
                return True
            elif status_code > 403:
                raise RemoteApiException("Failed to register installation to installation_auth/register", status_code)

            # Allow an extension for the session for 7 days without an internet connection
            if self.__created_older_than_x_days(days=7):
                # Reset the session - Unmanic should phone home once every 7 days
                self.__reset_session_installation_data()
            return False
        except RemoteApiException as e:
            self.logger.error("Exception while registering Unmanic: %s", e)
        except Exception as e:
            self.logger.debug("Exception while registering Unmanic: %s", e, exc_info=True)
            if self.__check_session_valid():
                # If the session is still valid, just return true. Perhaps the internet is down and it timed out?
                return True
            return False

    def sign_out(self):
        """
        Remove any user auth

        :return:
        """
        post_data = {
            "uuid": self.get_installation_uuid(),
        }
        response, status_code = self.api_post('unmanic-api', 1,
                                              'installation_auth/remove_installation_registration',
                                              post_data)
        # Save data
        self.logger.debug("Remote registry logout response - Code: %s, Body: %s", status_code, response)
        self.__reset_session_installation_data()
        return True

    def get_sign_out_url(self):
        """
        Fetch the application sign out URL

        :return:
        """
        return "{0}/unmanic-api/v1/installation_auth/logout".format(self.get_site_url())

    def get_patreon_login_url(self):
        """
        Fetch the Patreon Login URL

        :return:
        """
        return "{0}/support-auth-api/v1/login_patreon/login".format(self.get_site_url())

    def get_github_login_url(self):
        """
        Fetch the GitHub Login URL

        :return:
        """
        return "{0}/support-auth-api/v1/login_github/login".format(self.get_site_url())

    def get_discord_login_url(self):
        """
        Fetch the Discord Login URL

        :return:
        """
        return "{0}/support-auth-api/v1/login_discord/login".format(self.get_site_url())

    def get_patreon_sponsor_page(self):
        """
        Fetch the Patreon sponsor page

        :return:
        """
        try:
            # Fetch Patreon sponsorship URL from Unmanic site API
            response, status_code = self.api_get('unmanic-api', 1, 'links/unmanic_patreon_sponsor_page')
            if status_code in [200, 201, 202] and response.get("success"):
                response_data = response.get("data")
                return response_data
        except Exception as e:
            self.logger.debug('Exception while fetching Patreon sponsor page - %s', e)
        return False



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    unmanic.notifications.py

    Written by:               Josh.5 <jsunnex@gmail.com>
    Date:                     03 Jul 2022, (7:49 AM)

    Copyright:
           Copyright (C) Josh Sunnex - All Rights Reserved

           Permission is hereby granted, free of charge, to any person obtaining a copy
           of this software and associated documentation files (the "Software"), to deal
           in the Software without restriction, including without limitation the rights
           to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
           copies of the Software, and to permit persons to whom the Software is
           furnished to do so, subject to the following conditions:

           The above copyright notice and this permission notice shall be included in all
           copies or substantial portions of the Software.

           THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
           EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
           MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
           IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
           DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
           OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
           OR OTHER DEALINGS IN THE SOFTWARE.

"""
import threading
import uuid
from queue import Queue

from unmanic.libs.singleton import SingletonType


class Notifications(Queue, metaclass=SingletonType):
    """
    Handles messages passed to the frontend.

    Messages are sent as objects. These objects require the following fields:
        - 'uuid'            : A unique ID of the message. Prevent messages duplication
        - 'type'            : The type of message - 'error', 'warning', 'success', or 'info'
        - 'icon'            : The icon to display with the notification
        - 'label'           : A code to represent an I18n string for the frontend to display the notification label
        - 'message'         : The message string that is appended to the I18n string displayed on the frontend.
        - 'link'            : The link to be applied to the notification. Can be relative or a URL

    Examples:
        notifications = Notifications()

        # Place a notification
        notifications.add(
            {
                'uuid':       'failedTask',
                'type':       'error',
                'icon':       'report',
                'label':      'failedTaskLabel',
                'message':    'You have a new failed task in your completed tasks list',
                'navigation': {
                    'push': '/unmanic/ui/dashboard',
                    'events': [
                        'completedTasksShowMore',
                    ],
                },
            })

        # Remove an item
        notifications.remove('updateAvailable')

        # Read the current list of notifications
        notifications.read_all_items()

        # Update an item in place
        notifications.add(
            {
                'uuid':       'updateAvailable',
                'type':       'info',
                'icon':       'update',
                'label':      'updateAvailableLabel',
                'message':    'updateAvailableMessage',
                'navigation': {
                    'url': 'https://docs.unmanic.app',
                },
            })

    """

    def _init(self, maxsize):
        self.all_items = set()
        Queue._init(self, maxsize)

    def __add_to_queue(self, item, block=True, timeout=None):
        Queue.put(self, item, block, timeout)

    @staticmethod
    def __validate_item(item):
        # Ensure all required keys are present
        for key in ['type', 'icon', 'label', 'message', 'navigation']:
            if key not in item:
                raise Exception("Frontend message item incorrectly formatted. Missing key: '{}'".format(key))

        # Ensure the given type is valid
        if item.get('type') not in ['error', 'warning', 'success', 'info']:
            raise Exception(
                "Frontend message item's code must be in ['error', 'warning', 'success', 'info', 'status']. Received '{}'".format(
                    item.get('type')
                )
            )
        return True

    def __get_all_items(self):
        items = []
        while not self.empty():
            items.append(self.get())
        return items

    def __requeue_items(self, items):
        for item in items:
            self.__add_to_queue(item)

    def add(self, item):
        # Ensure received item is valid
        self.__validate_item(item)
        # Generate uuid if one is not provided
        if not item.get('uuid'):
            item['uuid'] = str(uuid.uuid4())
        # If it is not already in message list, add it to the list and the queue
        if item.get('uuid') not in self.all_items:
            self.all_items.add(item.get('uuid'))
            self.__add_to_queue(item)

    def remove(self, item_uuid):
        """
        Remove a single item from the notifications list given it's UUID

        :param item_uuid:
        :return:
        """
        success = False
        # Get all items out of queue
        current_items = self.__get_all_items()
        # Create list of items that will be queued again
        requeue_items = []
        for current_item in current_items:
            if current_item.get('uuid') != item_uuid:
                requeue_items.append(current_item)
        # Remove the requested item UUID from the all_items set
        lock = threading.RLock()
        lock.acquire()
        if item_uuid in self.all_items:
            self.all_items.remove(item_uuid)
            success = True
        lock.release()
        # Add all requeue_items items back into the queue
        self.__requeue_items(requeue_items)
        return success

    def read_all_items(self):
        # Get all items out of queue
        current_items = self.__get_all_items()
        # Add all requeue_items items back into the queue
        self.__requeue_items(current_items)
        # Return items list
        return current_items

    def update(self, item):
        # Ensure received item is valid
        self.__validate_item(item)
        # Generate uuid if one is not provided
        if not item.get('uuid'):
            item['uuid'] = str(uuid.uuid4())
        # If it is not already in message list, add it to the list and the queue
        if item.get('uuid') not in self.all_items:
            self.all_items.add(item.get('uuid'))
            self.__add_to_queue(item)
        else:
            # Get all items out of queue
            current_items = self.__get_all_items()
            # Create list of items that will be queued again
            # This will not include the item requested for update
            lock = threading.RLock()
            lock.acquire()
            requeue_items = []
            for current_item in current_items:
                if current_item.get('uuid') != item.get('uuid'):
                    requeue_items.append(current_item)
                    continue
                requeue_items.append(item)
            # Add all requeue_items items back into the queue
            self.__requeue_items(requeue_items)



"""Peewee migrations -- 001_rename_ffmpeg_log_to_log.py.

Some examples (model - class or model name)::

    > Model = migrator.orm['model_name']            # Return model in current state by name

    > migrator.sql(sql)                             # Run custom SQL
    > migrator.python(func, *args, **kwargs)        # Run python code
    > migrator.create_model(Model)                  # Create a model (could be used as decorator)
    > migrator.remove_model(model, cascade=True)    # Remove a model
    > migrator.add_fields(model, **fields)          # Add fields to a model
    > migrator.change_fields(model, **fields)       # Change fields
    > migrator.remove_fields(model, *field_names, cascade=True)
    > migrator.rename_field(model, old_field_name, new_field_name)
    > migrator.rename_table(model, new_table_name)
    > migrator.add_index(model, *col_names, unique=False)
    > migrator.drop_index(model, *col_names)
    > migrator.add_not_null(model, *field_names)
    > migrator.drop_not_null(model, *field_names)
    > migrator.add_default(model, field_name, default)

"""

import peewee as pw
from decimal import ROUND_HALF_EVEN

try:
    import playhouse.postgres_ext as pw_pext
except ImportError:
    pass

SQL = pw.SQL

"""NOTES:
The migrator function 'rename_field' has a bug: https://github.com/klen/peewee_migrate/issues/99
The simple work-around to this was to skip this method and just directly append a migration to the migrator's ops list.
Eg. `migrator.ops.append(migrator.migrator.rename_column('tasks', 'ffmpeg_log', 'log'))`
"""


"""Note:
This migration is required for legacy installations.
The old Unmanic had a ffmpeg_log column which had a NOT NULL contraint on it.
If this column is left as is, no items will be able to be added to the task queue.
"""
def migrate(migrator, database, fake=False, **kwargs):
    """Write your migrations here."""
    # Rename 'ffmpeg_log' field to 'log'' in Tasks model
    if any(cm for cm in database.get_columns('tasks') if cm.name == 'ffmpeg_log'):
        migrator.ops.append(migrator.migrator.rename_column('tasks', 'ffmpeg_log', 'log'))


def rollback(migrator, database, fake=False, **kwargs):
    """Write your rollback migrations here."""
    # Reverse rename 'ffmpeg_log' field to 'log'' in Tasks model
    if any(cm for cm in database.get_columns('tasks') if cm.name == 'log'):
        migrator.ops.append(migrator.migrator.rename_column('tasks', 'log', 'ffmpeg_log'))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    unmanic.plugins.py

    Written by:               Josh.5 <jsunnex@gmail.com>
    Date:                     01 Aug 2021, (9:35 AM)

    Copyright:
           Copyright (C) Josh Sunnex - All Rights Reserved

           Permission is hereby granted, free of charge, to any person obtaining a copy
           of this software and associated documentation files (the "Software"), to deal
           in the Software without restriction, including without limitation the rights
           to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
           copies of the Software, and to permit persons to whom the Software is
           furnished to do so, subject to the following conditions:

           The above copyright notice and this permission notice shall be included in all
           copies or substantial portions of the Software.

           THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
           EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
           MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
           IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
           DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
           OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
           OR OTHER DEALINGS IN THE SOFTWARE.

"""
import hashlib

from unmanic.libs.plugins import PluginsHandler
from unmanic.libs.unplugins import PluginExecutor


def prepare_filtered_plugins(params):
    """
    Returns a object of records filtered and sorted
    according to the provided request.

    :param params:
    :return:
    """
    start = params.get('start', 0)
    length = params.get('length', 0)

    search_value = params.get('search_value', '')

    # Note that plugins can be ordered in multiple ways. So this must be a list
    order = [
        params.get('order', {
            "column": 'name',
            "dir":    'desc',
        })
    ]

    # Fetch Plugins
    plugins = PluginsHandler()
    plugin_executor = PluginExecutor()
    # Get total count
    records_total_count = plugins.get_total_plugin_list_count()
    # Get quantity after filters (without pagination)
    records_filtered_count = plugins.get_plugin_list_filtered_and_sorted(order=order, start=0, length=0,
                                                                         search_value=search_value).count()
    # Get filtered/sorted results
    plugin_results = plugins.get_plugin_list_filtered_and_sorted(order=order, start=start, length=length,
                                                                 search_value=search_value)

    # Build return data
    return_data = {
        "recordsTotal":    records_total_count,
        "recordsFiltered": records_filtered_count,
        "results":         []
    }

    # Iterate over plugins and append them to the plugin data
    for plugin_result in plugin_results:
        # Set plugin status
        plugin_status = {
            "update_available": plugin_result.get('update_available'),
        }
        # Check if plugin is able to be configured
        has_config = False
        plugin_settings, plugin_settings_meta = plugin_executor.get_plugin_settings(plugin_result.get('plugin_id'))
        if plugin_settings:
            has_config = True
        # Set params as required in template
        item = {
            'id':          plugin_result.get('id'),
            'plugin_id':   plugin_result.get('plugin_id'),
            'icon':        plugin_result.get('icon'),
            'name':        plugin_result.get('name'),
            'description': plugin_result.get('description'),
            'tags':        plugin_result.get('tags'),
            'author':      plugin_result.get('author'),
            'version':     plugin_result.get('version'),
            'status':      plugin_status,
            'has_config':  has_config,
        }
        return_data["results"].append(item)

    # Return results
    return return_data


def get_plugin_types_with_flows():
    """
    Returns a list of all available plugin types

    :return:
    """
    return PluginsHandler.get_plugin_types_with_flows()


def get_enabled_plugin_flows_for_plugin_type(plugin_type, library_id):
    """
    Fetch all enabled plugin flows for a plugin type

    :param plugin_type:
    :param library_id:
    :return:
    """
    plugin_handler = PluginsHandler()
    return plugin_handler.get_enabled_plugin_flows_for_plugin_type(plugin_type, library_id)


def get_enabled_plugin_data_panels():
    """
    Returns a list of all enabled plugin data panels

    :return:
    """
    plugin_handler = PluginsHandler()
    return plugin_handler.get_enabled_plugin_modules_by_type('frontend.panel')


def exec_data_panels_plugin_runner(data, plugin_id):
    """
    Exec a frontend.panel plugin runner

    :param data:
    :param plugin_id:
    :return:
    """
    plugin_handler = PluginsHandler()
    return plugin_handler.exec_plugin_runner(data, plugin_id, 'frontend.panel')


def get_enabled_plugin_plugin_apis():
    """
    Returns a list of all enabled plugin APIs

    :return:
    """
    plugin_handler = PluginsHandler()
    return plugin_handler.get_enabled_plugin_modules_by_type('frontend.plugin_api')


def exec_plugin_api_plugin_runner(data, plugin_id):
    """
    Exec a frontend.plugin_api plugin runner

    :param data:
    :param plugin_id:
    :return:
    """
    plugin_handler = PluginsHandler()
    return plugin_handler.exec_plugin_runner(data, plugin_id, 'frontend.plugin_api')


def save_enabled_plugin_flows_for_plugin_type(plugin_type, library_id, plugin_flow):
    """
    Save a plugin flow given the plugin type and library ID

    :param plugin_type:
    :param library_id:
    :param plugin_flow:
    :return:
    """
    plugins = PluginsHandler()
    return plugins.set_plugin_flow(plugin_type, library_id, plugin_flow)


def remove_plugins(plugin_table_ids):
    """
    Removes/Uninstalls a list of plugins

    :param plugin_table_ids:
    :return:
    """
    plugins_handler = PluginsHandler()
    return plugins_handler.uninstall_plugins_by_db_table_id(plugin_table_ids)


def update_plugins(plugin_table_ids):
    """
    Removes/Uninstalls a list of plugins

    :param plugin_table_ids:
    :return:
    """
    plugins_handler = PluginsHandler()
    return plugins_handler.update_plugins_by_db_table_id(plugin_table_ids)


def get_plugin_settings(plugin_id: str, library_id=None):
    """
    Given a plugin installation ID, return a list of plugin settings for that plugin

    :param plugin_id:
    :param library_id:
    :return:
    """
    settings = []

    # Check plugin for settings
    plugin_executor = PluginExecutor()
    plugin_settings, plugin_settings_meta = plugin_executor.get_plugin_settings(plugin_id, library_id=library_id)
    if plugin_settings:
        for key in plugin_settings:
            form_input = {
                "key_id":         hashlib.md5(key.encode('utf8')).hexdigest(),
                "key":            key,
                "value":          plugin_settings.get(key),
                "input_type":     None,
                "label":          None,
                "description":    None,
                "tooltip":        None,
                "select_options": [],
                "slider_options": {},
                "display":        "visible",
                "sub_setting":    False,
            }

            plugin_setting_meta = plugin_settings_meta.get(key, {})

            # Set input type for form
            form_input['input_type'] = plugin_setting_meta.get('input_type', None)
            if not form_input['input_type']:
                form_input['input_type'] = "text"
                if isinstance(form_input['value'], bool):
                    form_input['input_type'] = "checkbox"

            # Handle unsupported input types (where they may be supported in future versions of Unmanic)
            supported_input_types = [
                "text",
                "textarea",
                "select",
                "checkbox",
                "slider",
                "browse_directory",
            ]
            if form_input['input_type'] not in supported_input_types:
                form_input['input_type'] = "text"

            # Set input display options
            form_input['display'] = plugin_setting_meta.get('display', 'visible')
            form_input['sub_setting'] = plugin_setting_meta.get('sub_setting', False)

            # Set input label text
            form_input['label'] = plugin_setting_meta.get('label', None)
            if not form_input['label']:
                form_input['label'] = key

            # Set input description text
            form_input['description'] = plugin_setting_meta.get('description', '')

            # Set input tooltip text
            form_input['tooltip'] = plugin_setting_meta.get('tooltip', '')

            # Set options if form input is select
            if form_input['input_type'] == 'select':
                form_input['select_options'] = plugin_setting_meta.get('select_options', [])
                if not form_input['select_options']:
                    # No options are given. Revert back to text input
                    form_input['input_type'] = 'text'

            # Set options if form input is slider
            if form_input['input_type'] == 'slider':
                slider_options = plugin_setting_meta.get('slider_options')
                if not slider_options:
                    # No options are given. Revert back to text input
                    form_input['input_type'] = 'text'
                else:
                    form_input['slider_options'] = {
                        'min':    slider_options.get('min', '0'),
                        'max':    slider_options.get('max', '1'),
                        'step':   slider_options.get('step', '1'),
                        'suffix': slider_options.get('suffix', ''),
                    }

            settings.append(form_input)
    return settings


def get_plugin_changelog(plugin_id):
    """
    Given a plugin installation ID, return a list of lines read from the plugin's changelog

    :param plugin_id:
    :return:
    """
    # Fetch plugin changelog
    plugin_executor = PluginExecutor()
    return plugin_executor.get_plugin_changelog(plugin_id)


def get_plugin_long_description(plugin_id):
    """
    Given a plugin installation ID, return a list of lines read from the plugin's changelog

    :param plugin_id:
    :return:
    """
    # Fetch plugin changelog
    plugin_executor = PluginExecutor()
    return plugin_executor.get_plugin_long_description(plugin_id)


def prepare_plugin_info_and_settings(plugin_id, prefer_local=True, library_id=None):
    """
    Returns a object of plugin metadata and current settings for the requested plugin_id

    :param prefer_local:
    :param plugin_id:
    :param library_id:
    :return:
    """
    plugins_handler = PluginsHandler()

    plugin_installed = True
    plugin_results = plugins_handler.get_plugin_list_filtered_and_sorted(plugin_id=plugin_id)

    if not plugin_results:
        # This plugin is not installed
        plugin_installed = False

    if not plugin_results or not prefer_local:
        # Try to fetch it from the repository
        plugin_list = plugins_handler.get_installable_plugins_list()
        for plugin in plugin_list:
            if plugin.get('plugin_id') == plugin_id:
                # Create changelog text from remote changelog text file
                plugin['changelog'] = plugins_handler.read_remote_changelog_file(plugin.get('changelog_url'))
                # Create list as the 'plugin_results' var above will also have returned a list if any results were found.
                plugin_results = [plugin]
                break

    # Iterate over plugins and append them to the plugin data
    plugin_data = {}
    for plugin_result in plugin_results:
        # Set plugin status
        plugin_status = {
            "installed":        plugin_result.get('installed', False),
            "update_available": plugin_result.get('update_available', False),
        }
        # Set params as required in template
        plugin_data = {
            'id':          plugin_result.get('id'),
            'plugin_id':   plugin_result.get('plugin_id'),
            'icon':        plugin_result.get('icon'),
            'name':        plugin_result.get('name'),
            'description': plugin_result.get('description'),
            'tags':        plugin_result.get('tags'),
            'author':      plugin_result.get('author'),
            'version':     plugin_result.get('version'),
            'changelog':   plugin_result.get('changelog', ''),
            'status':      plugin_status,
            'settings':    [],
        }
        if plugin_installed:
            plugin_data['settings'] = get_plugin_settings(plugin_result.get('plugin_id'), library_id=library_id)
            plugin_data['changelog'] = "".join(get_plugin_changelog(plugin_result.get('plugin_id')))
            plugin_data['description'] += "\n" + "".join(
                get_plugin_long_description(plugin_result.get('plugin_id')))
        break

    return plugin_data


def check_if_plugin_is_installed(plugin_id):
    """
    Returns true if the given plugin is installed

    :param plugin_id:
    :return:
    """
    plugins_handler = PluginsHandler()

    plugin_installed = True
    plugin_results = plugins_handler.get_plugin_list_filtered_and_sorted(plugin_id=plugin_id)

    if not plugin_results:
        # This plugin is not installed
        plugin_installed = False

    return plugin_installed


def update_plugin_settings(plugin_id, settings, library_id=None):
    """
    Updates the settings for the requested plugin_id

    :param plugin_id:
    :param settings:
    :param library_id:
    :return:
    """
    # Fetch plugin info (and settings if any)
    plugin_data = prepare_plugin_info_and_settings(plugin_id, library_id=library_id)

    # If no plugin data was found for the posted plugin table ID, then return a failure response
    if not plugin_data:
        return False

    # Loop over all plugin settings in order to find matches in the posted params
    settings_to_save = {}
    for s in settings:
        key = s.get('key')
        key_id = s.get('key_id')
        input_type = s.get('input_type')
        # Check if setting is in params
        value = s.get('value')
        # Check if value should be boolean
        if input_type == 'checkbox':
            if isinstance(value, str):
                value = True if value.lower() == 'true' else False
            elif isinstance(value, int):
                value = True if value > 0 else False
        # Add that to our dictionary of settings to save
        settings_to_save[key] = value

    # If we found settings that need to be saved, save them...
    if settings_to_save:
        plugin_executor = PluginExecutor()
        saved_all_settings = plugin_executor.save_plugin_settings(plugin_data.get('plugin_id'),
                                                                  settings_to_save,
                                                                  library_id=library_id)
        # If the save function was successful
        if saved_all_settings:
            # Update settings in plugin data that will be returned
            return True

    return False


def reset_plugin_settings(plugin_id, library_id=None):
    """
    Reset a plugin's settings back to defaults (or global config if a library ID is provided)

    :param plugin_id:
    :param library_id:
    :return:
    """
    # Fetch plugin info (and settings if any)
    plugin_data = prepare_plugin_info_and_settings(plugin_id, library_id=library_id)

    # If no plugin data was found for the posted plugin table ID, then return a failure response
    if not plugin_data:
        return False

    # Reset the plugin settings
    plugin_executor = PluginExecutor()
    return plugin_executor.reset_plugin_settings(plugin_data.get('plugin_id'), library_id=library_id)


def prepare_installable_plugins_list():
    """
    Return a list of plugins able to be installed.
    At the moment this does not employ any pagination. The lists are small enough
    that it can all be done frontend. However that may change in the future.

    :return:
    """
    plugins = PluginsHandler()
    # Fetch a list of plugin data cached locally
    return plugins.get_installable_plugins_list()


def install_plugin_by_id(plugin_id, repo_id=None):
    """
    Install a plugin given its Plugin ID

    :param plugin_id:
    :param repo_id:
    :return:
    """

    # Fetch a list of plugin data cached locally
    plugins = PluginsHandler()
    return plugins.install_plugin_by_id(plugin_id, repo_id)


def save_plugin_repos_list(repos_list):
    plugins = PluginsHandler()
    return plugins.set_plugin_repos(repos_list)


def prepare_plugin_repos_list():
    """
    Return a list of plugin repos available to download from

    :return:
    """
    return_repos = []

    plugins = PluginsHandler()
    # Fetch the data again from the database
    current_repos = plugins.get_plugin_repos()

    # Remove the default plugin repo from the list
    default_repo = plugins.get_default_repo()
    for repo in current_repos:
        if not repo.get("path").startswith(default_repo):
            return_repos.append(repo)

    # Append metadata from repo cache files
    for repo in return_repos:
        repo_path = repo.get('path')
        repo_id = plugins.get_plugin_repo_id(repo_path)
        repo_data = plugins.read_repo_data(repo_id)
        repo_metadata = repo_data.get('repo', {})
        repo['id'] = repo_metadata.get('id')
        repo['icon'] = repo_metadata.get('icon')
        repo['name'] = repo_metadata.get('name')

    return return_repos


def reload_plugin_repos_data():
    """
    Reloads all plugin repos data from the configured URL path

    :return:
    """
    plugins = PluginsHandler()
    return plugins.update_plugin_repos()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    unmanic.scheduler.py

    Written by:               Josh.5 <jsunnex@gmail.com>
    Date:                     11 Sep 2021, (11:15 AM)

    Copyright:
           Copyright (C) Josh Sunnex - All Rights Reserved

           Permission is hereby granted, free of charge, to any person obtaining a copy
           of this software and associated documentation files (the "Software"), to deal
           in the Software without restriction, including without limitation the rights
           to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
           copies of the Software, and to permit persons to whom the Software is
           furnished to do so, subject to the following conditions:

           The above copyright notice and this permission notice shall be included in all
           copies or substantial portions of the Software.

           THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
           EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
           MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
           IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
           DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
           OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
           OR OTHER DEALINGS IN THE SOFTWARE.

"""
import random
import threading
import time
from datetime import datetime, timedelta

import schedule

from unmanic import config
from unmanic.libs import common, task, unlogger
from unmanic.libs.installation_link import Links
from unmanic.libs.plugins import PluginsHandler
from unmanic.libs.session import Session


class ScheduledTasksManager(threading.Thread):
    """
    Manage any tasks that Unmanic needs to execute at regular intervals
    """

    def __init__(self, event):
        super(ScheduledTasksManager, self).__init__(name='ScheduledTasksManager')
        self.logger = None
        self.event = event
        self.abort_flag = threading.Event()
        self.abort_flag.clear()
        self.scheduler = schedule.Scheduler()
        self.force_local_worker_timer = 0

    def _log(self, message, message2='', level="info"):
        if not self.logger:
            unmanic_logging = unlogger.UnmanicLogger.__call__()
            self.logger = unmanic_logging.get_logger(self.name)
        message = common.format_message(message, message2)
        getattr(self.logger, level)(message)

    def stop(self):
        self.abort_flag.set()

    def run(self):
        self._log("Starting ScheduledTasks Monitor loop")

        # Create scheduled tasks
        # Check the session every 60 minutes
        self.scheduler.every(60).minutes.do(self.register_unmanic)
        # Run the plugin repo update every 3 hours
        self.scheduler.every(3).hours.do(self.plugin_repo_update)
        # Run the remote installation link update every 10 seconds
        self.scheduler.every(10).seconds.do(self.update_remote_installation_links)
        # Run the remote installation distributed worker counter sync every minute
        self.scheduler.every(1).minutes.do(self.set_worker_count_based_on_remote_installation_links)
        # Run a completed task cleanup every 60 minutes and on startup
        self.scheduler.every(12).hours.do(self.manage_completed_tasks)
        self.manage_completed_tasks()

        # Loop every 2 seconds to check if a task is due to be run
        while not self.abort_flag.is_set():
            self.event.wait(2)
            # Check if scheduled task is due
            self.scheduler.run_pending()

        # Clear any tasks and exit
        self.scheduler.clear()
        self._log("Leaving ScheduledTasks Monitor loop...")

    def register_unmanic(self):
        self._log("Updating session data")
        s = Session()
        s.register_unmanic(force=True)

    def plugin_repo_update(self):
        self._log("Checking for updates to plugin repos")
        plugin_handler = PluginsHandler()
        plugin_handler.update_plugin_repos()

    def update_remote_installation_links(self):
        # Don't log this as it will happen often
        links = Links()
        links.update_all_remote_installation_links()

    def set_worker_count_based_on_remote_installation_links(self):
        settings = config.Config()

        # Get local task count as int
        task_handler = task.Task()
        local_task_count = int(task_handler.get_total_task_list_count())

        # Get target count
        target_count = int(settings.get_distributed_worker_count_target())
        # # TODO: Check if we should be aiming for one less than the target
        # if target_count > 1:
        #     target_count -= 1

        linked_configs = []
        for local_config in settings.get_remote_installations():
            if local_config.get('enable_distributed_worker_count'):
                linked_configs.append(local_config)

        # If no remote links are configured, then return here
        if not linked_configs:
            return

        # There is a link config with distributed worker counts enabled
        self._log("Syncing distributed worker count for this installation")

        # Get total tasks count of pending tasks across all linked_configs
        total_tasks = local_task_count
        for linked_config in linked_configs:
            total_tasks += int(linked_config.get('task_count', 0))

        # From the counts fetched from all linked_configs, balance out the target count (including this installation)
        allocated_worker_count = 0
        for linked_config in linked_configs:
            if linked_config.get('task_count', 0) == 0:
                continue
            allocated_worker_count += round((int(linked_config.get('task_count', 0)) / total_tasks) * target_count)

        # Calculate worker count for local
        target_workers_for_this_installation = 0
        if local_task_count > 0:
            target_workers_for_this_installation = round((local_task_count / total_tasks) * target_count)

        # If the total allocated worker count is now above our target, set this installation back to 0
        if allocated_worker_count > target_count:
            target_workers_for_this_installation = 0

        # Every 10-12 minutes (make it random), give this installation at least 1 worker if it has pending tasks.
        #       This should cause the pending task queue to sit idle if there is only one task in the queue and it will provide
        #           rotation of workers when the pending task queue is close to the same.
        #       EG. If time now (seconds) > time last checked (seconds) + 10mins (600 seconds) + random seconds within 2mins
        time_now = time.time()
        time_to_next_force_local_worker = int(self.force_local_worker_timer + 600 + random.randrange(120))
        if time_now > time_to_next_force_local_worker:
            if (local_task_count > 1) and (target_workers_for_this_installation < 1):
                target_workers_for_this_installation = 1
                self.force_local_worker_timer = time_now

        self._log("Configuring worker count as {} for this installation".format(target_workers_for_this_installation))
        settings.set_config_item('number_of_workers', target_workers_for_this_installation, save_settings=True)

    def manage_completed_tasks(self):
        settings = config.Config()
        # Only run if configured to auto manage completed tasks
        if not settings.get_auto_manage_completed_tasks():
            return

        self._log("Running completed task cleanup for this installation")
        max_age_in_days = settings.get_max_age_of_completed_tasks()
        date_x_days_ago = datetime.now() - timedelta(days=int(max_age_in_days))
        before_time = date_x_days_ago.timestamp()

        task_success = True
        inc_status = 'successfully'
        if not settings.get_always_keep_failed_tasks():
            inc_status = 'successfully or failed'
            task_success = None

        # Fetch completed tasks
        from unmanic.libs import history
        history_logging = history.History()
        count = history_logging.get_historic_task_list_filtered_and_sorted(task_success=task_success,
                                                                           before_time=before_time).count()
        results = history_logging.get_historic_task_list_filtered_and_sorted(task_success=task_success,
                                                                             before_time=before_time)

        if count == 0:
            self._log("Found no {} completed tasks older than {} days".format(inc_status, max_age_in_days))
            return

        self._log(
            "Found {} {} completed tasks older than {} days that should be removed".format(count, inc_status, max_age_in_days))
        if not history_logging.delete_historic_tasks_recursively(results):
            self._log("Failed to delete {} {} completed tasks".format(count, inc_status), level='error')
            return

        self._log("Deleted {} {} completed tasks".format(count, inc_status))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    unmanic.installation_link.py

    Written by:               Josh.5 <jsunnex@gmail.com>
    Date:                     28 Oct 2021, (7:24 PM)

    Copyright:
           Copyright (C) Josh Sunnex - All Rights Reserved

           Permission is hereby granted, free of charge, to any person obtaining a copy
           of this software and associated documentation files (the "Software"), to deal
           in the Software without restriction, including without limitation the rights
           to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
           copies of the Software, and to permit persons to whom the Software is
           furnished to do so, subject to the following conditions:

           The above copyright notice and this permission notice shall be included in all
           copies or substantial portions of the Software.

           THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
           EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
           MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
           IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
           DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
           OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
           OR OTHER DEALINGS IN THE SOFTWARE.

"""
import json
import os.path
import queue
import threading
import time

import requests
from requests.auth import HTTPBasicAuth
from requests_toolbelt import MultipartEncoder

from unmanic import config
from unmanic.libs import common, session, task, unlogger
from unmanic.libs.library import Library
from unmanic.libs.session import Session
from unmanic.libs.singleton import SingletonType


class RequestHandler:

    def __init__(self, *args, **kwargs):
        self.auth = kwargs.get('auth', '')
        # Set username (could be passed in as None)
        self.username = ''
        if kwargs.get('username'):
            self.username = kwargs.get('username')
        # Set password (could be passed in as None)
        self.password = ''
        if kwargs.get('password'):
            self.password = kwargs.get('password')

    def __get_request_auth(self):
        request_auth = None
        if self.auth and self.auth.lower() == 'basic':
            request_auth = HTTPBasicAuth(self.username, self.password)
        return request_auth

    def get(self, url, **kwargs):
        return requests.get(url, auth=self.__get_request_auth(), **kwargs)

    def post(self, url, **kwargs):
        return requests.post(url, auth=self.__get_request_auth(), **kwargs)

    def delete(self, url, **kwargs):
        return requests.delete(url, auth=self.__get_request_auth(), **kwargs)


class Links(object, metaclass=SingletonType):
    _network_transfer_lock = {}

    def __init__(self, *args, **kwargs):
        self.settings = config.Config()
        self.session = session.Session()
        unmanic_logging = unlogger.UnmanicLogger.__call__()
        self.logger = unmanic_logging.get_logger(__class__.__name__)

    def _log(self, message, message2='', level="info"):
        message = common.format_message(message, message2)
        getattr(self.logger, level)(message)

    def __format_address(self, address: str):
        # Strip all whitespace
        address = address.strip()
        # Add http if it does not exist
        if not address.lower().startswith('http'):
            address = "http://{}".format(address)
        # Strip any trailing slashes
        address = address.rstrip('/')
        return address

    def __merge_config_dicts(self, config_dict, compare_dict):
        for key in config_dict.keys():
            if config_dict.get(key) != compare_dict.get(key) and compare_dict.get(key) is not None:
                # Apply the new value
                config_dict[key] = compare_dict.get(key)
                # Also flag the dict as updated
                config_dict['last_updated'] = time.time()

    def __generate_default_config(self, config_dict: dict):
        return {
            "address":                         config_dict.get('address', '???'),
            "auth":                            config_dict.get('auth', 'None'),
            "username":                        config_dict.get('username', ''),
            "password":                        config_dict.get('password', ''),
            "enable_receiving_tasks":          config_dict.get('enable_receiving_tasks', False),
            "enable_sending_tasks":            config_dict.get('enable_sending_tasks', False),
            "enable_task_preloading":          config_dict.get('enable_task_preloading', True),
            "preloading_count":                config_dict.get('preloading_count', 2),
            "enable_checksum_validation":      config_dict.get('enable_checksum_validation', False),
            "enable_config_missing_libraries": config_dict.get('enable_config_missing_libraries', False),
            "enable_distributed_worker_count": config_dict.get('enable_distributed_worker_count', False),
            "name":                            config_dict.get('name', '???'),
            "version":                         config_dict.get('version', '???'),
            "uuid":                            config_dict.get('uuid', '???'),
            "available":                       config_dict.get('available', False),
            "task_count":                      config_dict.get('task_count', 0),
            "last_updated":                    config_dict.get('last_updated', time.time()),
        }

    def acquire_network_transfer_lock(self, url, transfer_limit=1, lock_type='send'):
        """
        Limit transfers to each installation to 1 at a time

        :param url:
        :param transfer_limit:
        :param lock_type:
        :return:
        """
        time_now = time.time()
        lock = threading.RLock()
        # Limit maximum transfer limit to 5
        if transfer_limit > 5:
            transfer_limit = 5
        # Acquire a lock if one is available
        with lock:
            for tx_lock in range(transfer_limit):
                lock_key = "[{}-{}]-{}".format(lock_type, tx_lock, url)
                if self._network_transfer_lock.get(lock_key, {}).get('expires', 0) < time_now:
                    # Create new upload lock that will expire in 1 minute
                    self._network_transfer_lock[lock_key] = {
                        'expires': (time_now + 60),
                    }
                    # Return success
                    return lock_key
            # Failed to acquire network transfer lock
            return False

    def release_network_transfer_lock(self, lock_key):
        """
        Expire the transfer lock for the given lock_key

        :param lock_key:
        :return:
        """
        lock = threading.RLock()
        with lock:
            # Expire the lock for this address
            self._network_transfer_lock[lock_key] = {}
            return True

    def remote_api_get(self, remote_config: dict, endpoint: str, timeout=2):
        """
        GET to remote installation API

        :param remote_config:
        :param endpoint:
        :param timeout:
        :return:
        """
        request_handler = RequestHandler(
            auth=remote_config.get('auth'),
            username=remote_config.get('username'),
            password=remote_config.get('password'),
        )
        address = self.__format_address(remote_config.get('address'))
        url = "{}{}".format(address, endpoint)
        res = request_handler.get(url, timeout=timeout)
        if res.status_code == 200:
            return res.json()
        elif res.status_code in [400, 404, 405, 500]:
            json_data = res.json()
            self._log("Error while executing GET on remote installation API - {}. Message: '{}'".format(
                endpoint,
                json_data.get('error')),
                message2=json_data.get('traceback', []), level='error')
        return {}

    def remote_api_post(self, remote_config: dict, endpoint: str, data: dict, timeout=2):
        """
        POST to remote installation API

        :param remote_config:
        :param endpoint:
        :param data:
        :param timeout:
        :return:
        """
        request_handler = RequestHandler(
            auth=remote_config.get('auth'),
            username=remote_config.get('username'),
            password=remote_config.get('password'),
        )
        address = self.__format_address(remote_config.get('address'))
        url = "{}{}".format(address, endpoint)
        res = request_handler.post(url, json=data, timeout=timeout)
        if res.status_code == 200:
            return res.json()
        elif res.status_code in [400, 404, 405, 500]:
            json_data = res.json()
            self._log("Error while executing POST on remote installation API - {}. Message: '{}'".format(
                endpoint,
                json_data.get('error')),
                message2=json_data.get('traceback', []), level='error')
            return json_data
        return {}

    def remote_api_post_file(self, remote_config: dict, endpoint: str, path: str):
        """
        Send a file to the remote installation
        No timeout is set so the request will continue until closed

        :param remote_config:
        :param endpoint:
        :param path:
        :return:
        """
        request_handler = RequestHandler(
            auth=remote_config.get('auth'),
            username=remote_config.get('username'),
            password=remote_config.get('password'),
        )
        address = self.__format_address(remote_config.get('address'))
        url = "{}{}".format(address, endpoint)
        # NOTE: If you remove a content type from the upload (text/plain) the file upload fails
        # NOTE2: The 'ith open(path, "rb") as f' method reads the file into memory before uploading.
        #   This is slow and not ideal for devices with small amounts of ram.
        #   ```
        #       with open(path, "rb") as f:
        #           files = {"fileName": (os.path.basename(path), f, 'text/plain')}
        #           res = requests.post(url, files=files)
        #   ```
        m = MultipartEncoder(fields={'fileName': (os.path.basename(path), open(path, 'rb'), 'text/plain')})
        res = request_handler.post(url, data=m, headers={'Content-Type': m.content_type})
        if res.status_code == 200:
            return res.json()
        elif res.status_code in [400, 404, 405, 500]:
            json_data = res.json()
            self._log("Error while uploading file to remote installation API - {}. Message: '{}'".format(
                endpoint,
                json_data.get('error')),
                message2=json_data.get('traceback', []), level='error')
        return {}

    def remote_api_delete(self, remote_config: dict, endpoint: str, data: dict, timeout=2):
        """
        DELETE to remote installation API

        :param remote_config:
        :param endpoint:
        :param data:
        :param timeout:
        :return:
        """
        request_handler = RequestHandler(
            auth=remote_config.get('auth'),
            username=remote_config.get('username'),
            password=remote_config.get('password'),
        )
        address = self.__format_address(remote_config.get('address'))
        url = "{}{}".format(address, endpoint)
        res = request_handler.delete(url, json=data, timeout=timeout)
        if res.status_code == 200:
            return res.json()
        elif res.status_code in [400, 404, 405, 500]:
            json_data = res.json()
            self._log("Error while executing DELETE on remote installation API - {}. Message: '{}'".format(
                endpoint,
                json_data.get('error')),
                message2=json_data.get('traceback', []), level='error')
        return {}

    def remote_api_get_download(self, remote_config: dict, endpoint: str, path: str):
        """
        Download a file from a remote installation

        :param remote_config:
        :param endpoint:
        :param path:
        :return:
        """
        request_handler = RequestHandler(
            auth=remote_config.get('auth'),
            username=remote_config.get('username'),
            password=remote_config.get('password'),
        )
        address = self.__format_address(remote_config.get('address'))
        url = "{}{}".format(address, endpoint)
        with request_handler.get(url, stream=True) as r:
            r.raise_for_status()
            with open(path, 'wb') as f:
                for chunk in r.iter_content(chunk_size=None):
                    if chunk:
                        f.write(chunk)
        return True

    def validate_remote_installation(self, address: str, **kwargs):
        """
        Validate a remote Unmanic installation by requesting
        its system info and version

        :param address:
        :param username:
        :param password:
        :return:
        """
        address = self.__format_address(address)

        request_handler = RequestHandler(
            auth=kwargs.get('auth'),
            username=kwargs.get('username'),
            password=kwargs.get('password'),
        )

        # Fetch config
        url = "{}/unmanic/api/v2/settings/configuration".format(address)
        res = request_handler.get(url, timeout=2)
        if res.status_code != 200:
            if res.status_code in [400, 404, 405, 500]:
                json_data = res.json()
                self._log("Error while fetching remote installation config. Message: '{}'".format(json_data.get('error')),
                          message2=json_data.get('traceback', []), level='error')
            return {}
        system_configuration_data = res.json()

        # Fetch settings
        url = "{}/unmanic/api/v2/settings/read".format(address)
        res = request_handler.get(url, timeout=2)
        if res.status_code != 200:
            if res.status_code in [400, 404, 405, 500]:
                json_data = res.json()
                self._log("Error while fetching remote installation settings. Message: '{}'".format(json_data.get('error')),
                          message2=json_data.get('traceback', []), level='error')
            return {}
        settings_data = res.json()

        # Fetch version
        url = "{}/unmanic/api/v2/version/read".format(address)
        res = request_handler.get(url, timeout=2)
        if res.status_code != 200:
            if res.status_code in [400, 404, 405, 500]:
                json_data = res.json()
                self._log("Error while fetching remote installation version. Message: '{}'".format(json_data.get('error')),
                          message2=json_data.get('traceback', []), level='error')
            return {}
        version_data = res.json()

        # Fetch version
        url = "{}/unmanic/api/v2/session/state".format(address)
        res = request_handler.get(url, timeout=2)
        if res.status_code != 200:
            if res.status_code in [400, 404, 405, 500]:
                json_data = res.json()
                self._log(
                    "Error while fetching remote installation session state. Message: '{}'".format(json_data.get('error')),
                    message2=json_data.get('traceback', []), level='error')
            return {}
        session_data = res.json()

        # Fetch task count data
        data = {
            "start":  0,
            "length": 1
        }
        url = "{}/unmanic/api/v2/pending/tasks".format(address)
        res = request_handler.post(url, json=data, timeout=2)
        if res.status_code != 200:
            if res.status_code in [400, 404, 405, 500]:
                json_data = res.json()
                self._log(
                    "Error while fetching remote installation pending task list. Message: '{}'".format(json_data.get('error')),
                    message2=json_data.get('traceback', []), level='error')
            return {}
        tasks_data = res.json()

        return {
            'system_configuration': system_configuration_data.get('configuration'),
            'settings':             settings_data.get('settings'),
            'version':              version_data.get('version'),
            'session':              {
                "level":       session_data.get('level'),
                "picture_uri": session_data.get('picture_uri'),
                "name":        session_data.get('name'),
                "email":       session_data.get('email'),
                "uuid":        session_data.get('uuid'),
            },
            'task_count':           int(tasks_data.get('recordsTotal', 0))
        }

    def update_all_remote_installation_links(self):
        """
        Updates the link status and configuration of linked remote installations

        :return:
        """
        save_settings = False
        installation_id_list = []
        remote_installations = []
        distributed_worker_count_target = self.settings.get_distributed_worker_count_target()
        for local_config in self.settings.get_remote_installations():
            # Ensure address is not added twice by comparing installation IDs
            # Items matching these checks will be skipped over and will not be added to the installation list
            #   that will be re-saved
            if local_config.get('uuid') in installation_id_list and local_config.get('uuid', '???') != '???':
                # Do not update this installation. By doing this it will be removed from the list
                save_settings = True
                continue

            # Ensure the address is something valid
            if not local_config.get('address'):
                save_settings = True
                continue

            # Remove any entries that have an unknown address and uuid
            if local_config.get('address') == '???' and local_config.get('uuid') == '???':
                save_settings = True
                continue

            # Fetch updated data
            installation_data = None
            try:
                installation_data = self.validate_remote_installation(local_config.get('address'),
                                                                      auth=local_config.get('auth'),
                                                                      username=local_config.get('username'),
                                                                      password=local_config.get('password'))
            except Exception:
                pass

            # Generate updated configured values
            updated_config = self.__generate_default_config(local_config)
            updated_config["available"] = False
            if installation_data:
                # Mark the installation as available
                updated_config["available"] = True

                # Append the current task count
                updated_config["task_count"] = installation_data.get('task_count', 0)

                merge_dict = {
                    "name":    installation_data.get('settings', {}).get('installation_name'),
                    "version": installation_data.get('version'),
                    "uuid":    installation_data.get('session', {}).get('uuid'),
                }
                self.__merge_config_dicts(updated_config, merge_dict)

                # Fetch the corresponding remote configuration for this local installation
                remote_config = {}
                try:
                    remote_config = self.fetch_remote_installation_link_config_for_this(local_config)
                except requests.exceptions.Timeout:
                    self._log("Request to fetch remote installation config timed out", level='warning')
                    updated_config["available"] = False
                except requests.exceptions.RequestException as e:
                    self._log("Request to fetch remote installation config failed", message2=str(e), level='warning')
                    updated_config["available"] = False
                except Exception as e:
                    self._log("Failed to fetch remote installation config", message2=str(e), level='error')
                    updated_config["available"] = False

                # If the remote configuration is newer than this one, use those values
                # The remote installation will do the same and this will synchronise
                remote_link_config = remote_config.get('link_config', {})
                if local_config.get('last_updated', 1) < remote_link_config.get('last_updated', 1):
                    # Note that the configuration options are reversed when reading from the remote installation config
                    # These items are not synced here:
                    #   - enable_task_preloading
                    #   - enable_checksum_validation
                    #   - enable_config_missing_libraries
                    if updated_config["enable_receiving_tasks"] != remote_link_config.get('enable_sending_tasks'):
                        updated_config["enable_receiving_tasks"] = remote_link_config.get('enable_sending_tasks')
                        save_settings = True
                    if updated_config["enable_sending_tasks"] != remote_link_config.get('enable_receiving_tasks'):
                        updated_config["enable_sending_tasks"] = remote_link_config.get('enable_receiving_tasks')
                        save_settings = True
                    # Update the distributed_worker_count_target
                    distributed_worker_count_target = remote_config.get('distributed_worker_count_target', 0)
                    # Also sync the last_updated flag
                    updated_config['last_updated'] = remote_link_config.get('last_updated')

                # If the remote config is unable to contact this installation (or it does not have a corresponding config yet)
                #   then also push the configuration
                if not remote_link_config.get('available'):
                    try:
                        self.push_remote_installation_link_config(updated_config)
                    except requests.exceptions.Timeout:
                        self._log("Request to push link config to remote installation timed out", level='warning')
                        updated_config["available"] = False
                    except requests.exceptions.RequestException as e:
                        self._log("Request to push link config to remote installation failed", message2=str(e),
                                  level='warning')
                        updated_config["available"] = False
                    except Exception as e:
                        self._log("Failed to push link config to remote installation", message2=str(e), level='error')
                        updated_config["available"] = False

                # Push library configurations for missing remote libraries (if configured to do so)
                if local_config.get('enable_sending_tasks') and local_config.get('enable_config_missing_libraries'):
                    # Fetch remote installation library name list
                    results = self.remote_api_get(local_config, '/unmanic/api/v2/settings/libraries')
                    existing_library_names = []
                    for library in results.get('libraries', []):
                        existing_library_names.append(library.get('name'))
                    # Loop over local libraries and create an import object for each one that is missing
                    for library in Library.get_all_libraries():
                        # Ignore local libraries that are configured for remote only
                        if library.get('enable_remote_only'):
                            continue
                        # For each of the missing libraries, create a new remote library with that config.
                        if library.get('name') not in existing_library_names:
                            # Export library config
                            import_data = Library.export(library.get('id'))
                            # Set library ID to 0 to generate new library from this import
                            import_data['library_id'] = 0
                            # Configure remote library to be fore remote files only
                            import_data['library_config']['enable_remote_only'] = True
                            import_data['library_config']['enable_scanner'] = False
                            import_data['library_config']['enable_inotify'] = False
                            # Import library on remote installation
                            self._log("Importing remote library config '{}'".format(library.get('name')), message2=import_data,
                                      level='debug')
                            result = self.import_remote_library_config(local_config, import_data)
                            if result is None:
                                # There was a connection issue of some kind. This was already logged.
                                continue
                            if result.get('success'):
                                self._log("Successfully imported library '{}'".format(library.get('name')), level='debug')
                                continue
                            self._log("Failed to import library config '{}'".format(library.get('name')),
                                      message2=result.get('error'), level='error')

            # Only save to file if the settings have been updated
            remote_installations.append(updated_config)

            # Add UUID to list for next loop
            installation_id_list.append(updated_config.get('uuid', '???'))

        # Update installation data. Only save the config to disk if it was modified
        settings_dict = {
            'remote_installations':            remote_installations,
            'distributed_worker_count_target': distributed_worker_count_target
        }
        self.settings.set_bulk_config_items(settings_dict, save_settings=save_settings)

        return remote_installations

    def read_remote_installation_link_config(self, uuid: str):
        """
        Returns the configuration of the remote installation

        :param uuid:
        :return:
        """
        for remote_installation in self.settings.get_remote_installations():
            if remote_installation.get('uuid') == uuid:
                # If not yet configured, set default values before returning
                return self.__generate_default_config(remote_installation)

        # Ensure we have settings data from the remote installation
        raise Exception("Unable to read installation link configuration.")

    def update_single_remote_installation_link_config(self, configuration: dict, distributed_worker_count_target=0):
        """
        Returns the configuration of the remote installation

        :param configuration:
        :param distributed_worker_count_target:
        :return:
        """
        uuid = configuration.get('uuid')
        if not uuid:
            raise Exception("Updating a single installation link configuration requires a UUID.")

        current_distributed_worker_count_target = self.settings.get_distributed_worker_count_target()
        force_update_flag = False
        if int(current_distributed_worker_count_target) != int(distributed_worker_count_target):
            force_update_flag = True

        config_exists = False
        remote_installations = []
        for local_config in self.settings.get_remote_installations():
            updated_config = self.__generate_default_config(local_config)

            # If this is the uuid in the config provided, then update our config with the provided values
            if local_config.get('uuid') == uuid:
                config_exists = True
                self.__merge_config_dicts(updated_config, configuration)

            # If this link is configured for distributed worker count, and that count was change,
            #   force the last update flag to be updated so this change is disseminated
            if force_update_flag and configuration.get('enable_distributed_worker_count'):
                updated_config['last_updated'] = time.time()

            remote_installations.append(updated_config)

        # If the config does not yet exist, the add it now
        if not config_exists:
            remote_installations.append(self.__generate_default_config(configuration))

        # Update installation data and save the config to disk
        settings_dict = {
            'remote_installations':            remote_installations,
            'distributed_worker_count_target': distributed_worker_count_target
        }
        self.settings.set_bulk_config_items(settings_dict, save_settings=True)

    def delete_remote_installation_link_config(self, uuid: str):
        """
        Removes a link configuration for a remote installation given its uuid
        If no uuid match is found, returns False

        :param uuid:
        :return:
        """
        removed = False
        updated_list = []
        for remote_installation in self.settings.get_remote_installations():
            if remote_installation.get('uuid') == uuid:
                # Mark the task as having successfully remoted the installation
                removed = True
                continue
            # Only add remote installations that do not match
            updated_list.append(remote_installation)

        # Update installation data and save the config to disk
        settings_dict = {
            'remote_installations': updated_list,
        }
        self.settings.set_bulk_config_items(settings_dict, save_settings=True)
        return removed

    def fetch_remote_installation_link_config_for_this(self, remote_config: dict):
        """
        Fetches and returns the corresponding link configuration from a remote installation

        :param remote_config:
        :return:
        """
        request_handler = RequestHandler(
            auth=remote_config.get('auth'),
            username=remote_config.get('username'),
            password=remote_config.get('password'),
        )
        address = self.__format_address(remote_config.get('address'))
        url = "{}/unmanic/api/v2/settings/link/read".format(address)
        data = {
            "uuid": self.session.uuid
        }
        res = request_handler.post(url, json=data, timeout=2)
        if res.status_code == 200:
            return res.json()
        elif res.status_code in [400, 404, 405, 500]:
            json_data = res.json()
            self._log("Error while fetching remote installation link config. Message: '{}'".format(json_data.get('error')),
                      message2=json_data.get('traceback', []), level='error')
        return {}

    def push_remote_installation_link_config(self, configuration: dict):
        """
        Pushes the given link config to the remote installation returns the corresponding link configuration from a remote installation

        :param configuration:
        :return:
        """
        request_handler = RequestHandler(
            auth=configuration.get('auth'),
            username=configuration.get('username'),
            password=configuration.get('password'),
        )
        address = self.__format_address(configuration.get('address'))
        url = "{}/unmanic/api/v2/settings/link/write".format(address)

        # First generate an updated config
        updated_config = self.__generate_default_config(configuration)

        # Update the bits for the remote instance
        updated_config['uuid'] = self.session.uuid
        updated_config['name'] = self.settings.get_installation_name()
        updated_config['version'] = self.settings.read_version()

        # Configure settings
        updated_config["enable_receiving_tasks"] = configuration.get('enable_sending_tasks')
        updated_config["enable_sending_tasks"] = configuration.get('enable_receiving_tasks')

        # Current task count
        task_handler = task.Task()
        updated_config["task_count"] = int(task_handler.get_total_task_list_count())

        # Fetch local config for distributed_worker_count_target
        distributed_worker_count_target = self.settings.get_distributed_worker_count_target()

        # Remove some of the other fields. These will need to be adjusted on the remote instance manually
        del updated_config['address']
        del updated_config['available']

        data = {
            'link_config':                     updated_config,
            'distributed_worker_count_target': distributed_worker_count_target
        }
        res = request_handler.post(url, json=data, timeout=2)
        if res.status_code == 200:
            return True
        elif res.status_code in [400, 404, 405, 500]:
            json_data = res.json()
            self._log("Error while pushing remote installation link config. Message: '{}'".format(json_data.get('error')),
                      message2=json_data.get('traceback', []), level='error')
        return False

    def check_remote_installation_for_available_workers(self):
        """
        Return a list of installations with workers available for a remote task.
        This list is filtered by:
            - Only installations that are available
            - Only installations that are configured for sending tasks to
            - Only installations that have not pending tasks
            - Only installations that have at least one idle worker that is not paused

        :return:
        """
        installations_with_info = {}
        for lc in self.settings.get_remote_installations():
            local_config = self.__generate_default_config(lc)

            # Only installations that are available
            if not local_config.get('available'):
                continue

            # Only installations that are configured for sending tasks to
            if not local_config.get('enable_sending_tasks'):
                continue

            # No valid UUID, no valid connection. This link may still be syncing
            if len(local_config.get('uuid', '')) < 20:
                continue

            try:
                # Define auth
                # Only installations that have at least one idle worker that is not paused
                results = self.remote_api_get(local_config, '/unmanic/api/v2/workers/status')
                worker_list = results.get('workers_status', [])

                # Only add installations that have not got pending tasks. This is unless we are configured to preload the queue
                max_pending_tasks = 0
                if local_config.get('enable_task_preloading'):
                    # Preload with the number of workers (regardless of the worker status) plus an additional one to account
                    # for delays in the downloads
                    max_pending_tasks = local_config.get('preloading_count')
                results = self.remote_api_post(local_config, '/unmanic/api/v2/pending/tasks', {
                    "start":  0,
                    "length": 1
                })
                if results.get('error'):
                    continue
                current_pending_tasks = int(results.get('recordsFiltered', 0))
                if local_config.get('enable_task_preloading') and current_pending_tasks >= max_pending_tasks:
                    self._log("Remote installation has exceeded the max remote pending task count ({})".format(
                        current_pending_tasks), level='debug')
                    continue

                # Fetch remote installation library name list
                results = self.remote_api_get(local_config, '/unmanic/api/v2/settings/libraries')
                library_names = []
                for library in results.get('libraries', []):
                    library_names.append(library.get('name'))

                # Ensure that worker count is more than 0
                if len(worker_list):
                    installations_with_info[local_config.get('uuid')] = {
                        "address":                local_config.get('address'),
                        "auth":                   local_config.get('auth'),
                        "username":               local_config.get('username'),
                        "password":               local_config.get('password'),
                        "enable_task_preloading": local_config.get('enable_task_preloading'),
                        "preloading_count":       local_config.get('preloading_count'),
                        "library_names":          library_names,
                        "available_slots":        0,
                    }

                available_workers = False
                for worker in worker_list:
                    # Add a slot for each worker regardless of its status
                    installations_with_info[local_config.get('uuid')]['available_slots'] += 1
                    if worker.get('idle') and not worker.get('paused'):
                        # If any workers are idle and not paused then we have an available worker slot
                        available_workers = True
                        installations_with_info[local_config.get('uuid')]['available_workers'] = True
                    elif not worker.get('idle'):
                        # If any workers are busy with a task then also mark that as an an available worker slot
                        available_workers = True
                        installations_with_info[local_config.get('uuid')]['available_workers'] = True

                # Check if this installation is configured for preloading
                if available_workers and local_config.get('enable_task_preloading'):
                    # Add more slots to fill up the pending task queue
                    while not current_pending_tasks > max_pending_tasks:
                        installations_with_info[local_config.get('uuid')]['available_slots'] += 1
                        current_pending_tasks += 1

            except Exception as e:
                self._log("Failed to contact remote installation '{}'".format(local_config.get('address')), message2=str(e),
                          level='warning')
                continue

        return installations_with_info

    def within_enabled_link_limits(self, frontend_messages=None):
        """
        Ensure enabled plugins are within limits

        :param frontend_messages:
        :return:
        """
        # Fetch level from session
        s = Session()
        s.register_unmanic()
        if s.level > 1:
            return True

        # Fetch all linked remote installations
        remote_installations = self.settings.get_remote_installations()

        def add_frontend_message():
            # If the frontend messages queue was included in request, append a message
            if frontend_messages:
                frontend_messages.put(
                    {
                        'id':      'linkedInstallationLimits',
                        'type':    'error',
                        'code':    'linkedInstallationLimits',
                        'message': '',
                        'timeout': 0
                    }
                )

        # Ensure remote installations are within limits
        # Function was returned above if the user was logged in and able to use infinite
        if len(remote_installations) > s.link_count:
            add_frontend_message()
            return False
        return True

    def new_pending_task_create_on_remote_installation(self, remote_config: dict, abspath: str, library_id: int):
        """
        Create a new pending task on a remote installation.
        The remote installation will return the ID of a generated task.

        :param remote_config:
        :param abspath:
        :param library_id:
        :return:
        """
        try:
            request_handler = RequestHandler(
                auth=remote_config.get('auth'),
                username=remote_config.get('username'),
                password=remote_config.get('password'),
            )
            address = self.__format_address(remote_config.get('address'))
            url = "{}/unmanic/api/v2/pending/create".format(address)
            data = {
                "path":       abspath,
                "library_id": library_id,
                "type":       'remote',
            }
            res = request_handler.post(url, json=data, timeout=2)
            if res.status_code in [200, 400]:
                return res.json()
            elif res.status_code in [404, 405, 500]:
                json_data = res.json()
                self._log("Error while creating new remote pending task. Message: '{}'".format(json_data.get('error')),
                          message2=json_data.get('traceback', []), level='error')
            return {}
        except requests.exceptions.Timeout:
            self._log("Request to create remote pending task timed out '{}'".format(abspath), level='warning')
            return None
        except requests.exceptions.RequestException as e:
            self._log("Request to create remote pending task failed '{}'".format(abspath), message2=str(e), level='warning')
            return None
        except Exception as e:
            self._log("Failed to create remote pending task '{}'".format(abspath), message2=str(e), level='error')
        return {}

    def send_file_to_remote_installation(self, remote_config: dict, path: str):
        """
        Send a file to a remote installation.
        The remote installation will return the ID of a generated task.

        :param remote_config:
        :param path:
        :return:
        """
        try:
            results = self.remote_api_post_file(remote_config, '/unmanic/api/v2/upload/pending/file', path)
            if results.get('error'):
                results = {}
            return results
        except requests.exceptions.RequestException as e:
            self._log("Request to upload to remote installation failed", message2=str(e), level='warning')
        except Exception as e:
            self._log("Failed to upload to remote installation", message2=str(e), level='error')
        return {}

    def remove_task_from_remote_installation(self, remote_config: dict, remote_task_id: int):
        """
        Remove a task from the pending queue

        :param remote_config:
        :param remote_task_id:
        :return:
        """
        try:
            data = {
                "id_list": [remote_task_id]
            }
            return self.remote_api_delete(remote_config, '/unmanic/api/v2/pending/tasks', data, timeout=15)
        except requests.exceptions.Timeout:
            self._log("Request to remove remote task timed out", level='warning')
            return None
        except requests.exceptions.RequestException as e:
            self._log("Request to remove remote task failed", message2=str(e), level='warning')
            return None
        except Exception as e:
            self._log("Failed to remove remote pending task", message2=str(e), level='error')
        return {}

    def get_the_remote_library_config_by_name(self, remote_config: dict, library_name: str):
        """
        Fetch a remote library's configuration by its name

        :param remote_config:
        :param library_name:
        :return:
        """
        try:
            # Fetch remote installation libraries
            results = self.remote_api_get(remote_config, '/unmanic/api/v2/settings/libraries', timeout=4)
            for library in results.get('libraries', []):
                if library.get('name') == library_name:
                    return library
        except requests.exceptions.Timeout:
            self._log("Request to set remote task library timed out", level='warning')
            return None
        except requests.exceptions.RequestException as e:
            self._log("Request to set remote task library failed", message2=str(e), level='warning')
            return None
        except Exception as e:
            self._log("Failed to set remote task library", message2=str(e), level='error')
        return {}

    def set_the_remote_task_library(self, remote_config: dict, remote_task_id: int, library_name: str):
        """
        Set the library for the remote task
        Defaults to the remote installation's default library

        :param remote_config:
        :param remote_task_id:
        :param library_name:
        :return:
        """
        try:
            data = {
                "id_list":      [remote_task_id],
                "library_name": library_name,
            }
            results = self.remote_api_post(remote_config, '/unmanic/api/v2/pending/library/update', data, timeout=7)
            if results.get('error'):
                results = {}
            return results
        except requests.exceptions.Timeout:
            self._log("Request to set remote task library timed out", level='warning')
            return None
        except requests.exceptions.RequestException as e:
            self._log("Request to set remote task library failed", message2=str(e), level='warning')
            return None
        except Exception as e:
            self._log("Failed to set remote task library", message2=str(e), level='error')
        return {}

    def get_remote_pending_task_state(self, remote_config: dict, remote_task_id: int):
        """
        Get the remote pending task status

        :param remote_config:
        :param remote_task_id:
        :return:
        """
        try:
            data = {
                "id_list": [remote_task_id]
            }
            results = self.remote_api_post(remote_config, '/unmanic/api/v2/pending/status/get', data, timeout=7)
            return results
        except requests.exceptions.Timeout:
            self._log("Request to get status of remote task timed out", level='warning')
        except requests.exceptions.RequestException as e:
            self._log("Request to get status of remote task failed", message2=str(e), level='warning')
        except Exception as e:
            self._log("Failed to get status of remote pending task", message2=str(e), level='error')
        return None

    def start_the_remote_task_by_id(self, remote_config: dict, remote_task_id: int):
        """
        Start the remote pending task

        :param remote_config:
        :param remote_task_id:
        :return:
        """
        try:
            data = {
                "id_list": [remote_task_id]
            }
            results = self.remote_api_post(remote_config, '/unmanic/api/v2/pending/status/set/ready', data, timeout=7)
            if results.get('error'):
                results = {}
            return results
        except requests.exceptions.Timeout:
            self._log("Request to start remote task timed out", level='warning')
            return None
        except requests.exceptions.RequestException as e:
            self._log("Request to start remote task failed", message2=str(e), level='warning')
            return None
        except Exception as e:
            self._log("Failed to start remote pending task", message2=str(e), level='error')
        return {}

    def get_all_worker_status(self, remote_config: dict):
        """
        Start the remote pending task

        :param remote_config:
        :return:
        """
        try:
            results = self.remote_api_get(remote_config, '/unmanic/api/v2/workers/status')
            return results.get('workers_status', [])
        except requests.exceptions.Timeout:
            self._log("Request to get worker status timed out", level='warning')
        except requests.exceptions.RequestException as e:
            self._log("Request to get worker status failed", message2=str(e), level='warning')
        except Exception as e:
            self._log("Failed to get worker status", message2=str(e), level='error')
        return []

    def get_single_worker_status(self, remote_config: dict, worker_id: str):
        """
        Start the remote pending task

        :param remote_config:
        :param worker_id:
        :return:
        """
        workers_status = self.get_all_worker_status(remote_config)
        for worker in workers_status:
            if worker.get('id') == worker_id:
                return worker
        return {}

    def terminate_remote_worker(self, remote_config: dict, worker_id: str):
        """
        Start the remote pending task

        :param remote_config:
        :param worker_id:
        :return:
        """
        try:
            data = {
                "worker_id": [worker_id]
            }
            return self.remote_api_delete(remote_config, '/unmanic/api/v2/workers/worker/terminate', data)
        except requests.exceptions.Timeout:
            self._log("Request to terminate remote worker timed out", level='warning')
        except requests.exceptions.RequestException as e:
            self._log("Request to terminate remote worker failed", message2=str(e), level='warning')
        except Exception as e:
            self._log("Failed to terminate remote worker", message2=str(e), level='error')
        return {}

    def fetch_remote_task_data(self, remote_config: dict, remote_task_id: int, path: str):
        """
        Fetch the completed remote task data

        :param remote_config:
        :param remote_task_id:
        :param path:
        :return:
        """
        task_data = {}
        try:
            # Request API generate a DL link
            link_info = self.remote_api_get(remote_config,
                                            '/unmanic/api/v2/pending/download/data/id/{}'.format(remote_task_id))
            if link_info.get('link_id'):
                # Download the data file
                res = self.remote_api_get_download(remote_config, '/unmanic/downloads/{}'.format(link_info.get('link_id')),
                                                   path)
                if res and os.path.exists(path):
                    with open(path) as f:
                        task_data = json.load(f)
        except requests.exceptions.Timeout:
            self._log("Request to fetch remote task data timed out", level='warning')
        except requests.exceptions.RequestException as e:
            self._log("Request to fetch remote task data failed", message2=str(e), level='warning')
        except Exception as e:
            self._log("Failed to fetch remote task data", message2=str(e), level='error')
        return task_data

    def fetch_remote_task_completed_file(self, remote_config: dict, remote_task_id: int, path: str):
        """
        Fetch the completed remote task file

        :param remote_config:
        :param remote_task_id:
        :param path:
        :return:
        """
        try:
            # Request API generate a DL link
            link_info = self.remote_api_get(remote_config,
                                            '/unmanic/api/v2/pending/download/file/id/{}'.format(remote_task_id))
            if link_info.get('link_id'):
                # Download the file
                res = self.remote_api_get_download(remote_config, '/unmanic/downloads/{}'.format(link_info.get('link_id')),
                                                   path)
                if res and os.path.exists(path):
                    return True
        except requests.exceptions.Timeout:
            self._log("Request to fetch remote task completed file timed out", level='warning')
        except requests.exceptions.RequestException as e:
            self._log("Request to fetch remote task completed file failed", message2=str(e), level='warning')
        except Exception as e:
            self._log("Failed to fetch remote task completed file", message2=str(e), level='error')
        return False

    def import_remote_library_config(self, remote_config: dict, import_data: dict):
        """
        Import a library config on a remote installation

        :param remote_config:
        :param import_data:
        :return:
        """
        try:
            results = self.remote_api_post(remote_config, '/unmanic/api/v2/settings/library/import', import_data, timeout=60)
            if results.get('error'):
                results = {}
            return results
        except requests.exceptions.Timeout:
            self._log("Request to import remote library timed out", level='warning')
            return None
        except requests.exceptions.RequestException as e:
            self._log("Request to import remote library failed", message2=str(e), level='warning')
            return None
        except Exception as e:
            self._log("Failed to import remote library", message2=str(e), level='error')
        return {}


class RemoteTaskManager(threading.Thread):
    paused = False

    current_task = None
    worker_log = None
    start_time = None
    finish_time = None

    worker_subprocess_percent = None
    worker_subprocess_elapsed = None

    worker_runners_info = {}

    def __init__(self, thread_id, name, installation_info, pending_queue, complete_queue, event):
        super(RemoteTaskManager, self).__init__(name=name)
        self.thread_id = thread_id
        self.name = name
        self.event = event
        self.installation_info = installation_info
        self.pending_queue = pending_queue
        self.complete_queue = complete_queue

        self.links = Links()

        # Create 'redundancy' flag. When this is set, the worker should die
        self.redundant_flag = threading.Event()
        self.redundant_flag.clear()

        # Create 'paused' flag. When this is set, the worker should be paused
        self.paused_flag = threading.Event()
        self.paused_flag.clear()

        # Create logger for this worker
        unmanic_logging = unlogger.UnmanicLogger.__call__()
        self.logger = unmanic_logging.get_logger(self.name)

    def _log(self, message, message2='', level="info"):
        message = common.format_message(message, message2)
        getattr(self.logger, level)(message)

    def get_info(self):
        return {
            'name':              self.name,
            'installation_info': self.installation_info,
        }

    def run(self):
        # A manager should only run for a single task and connection to a single worker.
        # If either of these become unavailable, then the manager should exit
        self._log("Starting remote task manager {} - {}".format(self.thread_id, self.installation_info.get('address')))
        # Pull task
        try:
            # Pending task queue has an item available. Fetch it.
            next_task = self.pending_queue.get_nowait()

            # Configure worker for this task
            self.__set_current_task(next_task)

            # Process the set task
            self.__process_task_queue_item()

        except queue.Empty:
            self._log("Remote task manager started by the pending queue was empty", level="warning")
        except Exception as e:
            self._log("Exception in processing job with {}:".format(self.name), message2=str(e),
                      level="exception")

        self._log("Stopping remote task manager {} - {}".format(self.thread_id, self.installation_info.get('address')))

    def __set_current_task(self, current_task):
        """Sets the given task to the worker class"""
        self.current_task = current_task
        self.worker_log = []

    def __unset_current_task(self):
        self.current_task = None
        self.worker_runners_info = {}
        self.worker_log = []

    def __process_task_queue_item(self):
        """
        Processes the set task.

        :return:
        """
        # Set the progress to an empty string
        self.worker_subprocess_percent = ''
        self.worker_subprocess_elapsed = '0'

        # Log the start of the job
        self._log("Picked up job - {}".format(self.current_task.get_source_abspath()))

        # Mark as being "in progress"
        self.current_task.set_status('in_progress')

        # Start current task stats
        self.__set_start_task_stats()

        # Process the file. Will return true if success, otherwise false
        success = self.__send_task_to_remote_worker_and_monitor()
        # Mark the task as either success or not
        self.current_task.set_success(success)

        # Mark task completion statistics
        self.__set_finish_task_stats()

        # Log completion of job
        self._log("Finished job - {}".format(self.current_task.get_source_abspath()))

        # Place the task into the completed queue
        self.complete_queue.put(self.current_task)

        # Reset the current file info for the next task
        self.__unset_current_task()

    def __set_start_task_stats(self):
        """Sets the initial stats for the start of a task"""
        # Set the start time to now
        self.start_time = time.time()

        # Clear the finish time
        self.finish_time = None

        # Format our starting statistics data
        self.current_task.task.processed_by_worker = self.name
        self.current_task.task.start_time = self.start_time
        self.current_task.task.finish_time = self.finish_time

    def __set_finish_task_stats(self):
        """Sets the final stats for the end of a task"""
        # Set the finish time to now
        self.finish_time = time.time()

        # Set the finish time in the statistics data
        self.current_task.task.finish_time = self.finish_time

    def __write_failure_to_worker_log(self):
        # Append long entry to say the worker was terminated
        self.worker_log.append("\n\nREMOTE TASK FAILED!")
        self.worker_log.append("\nAn error occurred during one of these stages:")
        self.worker_log.append("\n    - while sending task to remote installation")
        self.worker_log.append("\n    - during the remote task processing")
        self.worker_log.append("\n    - while attempting to retrieve the completed task from the remote installation")
        self.worker_log.append("\nCheck Unmanic logs for more information.")
        self.worker_log.append("\nRelevant logs will be prefixed with 'ERROR:Unmanic.{}'".format(self.name))
        self.current_task.save_command_log(self.worker_log)

    def __send_task_to_remote_worker_and_monitor(self):
        """
        Sends the task file to the remote installation to process.
        Monitors progress and then fetches the results

        TODO: Manage network disconnections.
            - This manager object should be able to handle a network disconnect. However, we should terminate
            this manager if the remote task no longer exists.
            - Catch all API request exceptions.
            - Remove the failed_status_count - losing contact should be ok. What matters is when contact is made that
            the task still exists to be downloaded or status updated.

        :return:
        """
        # Set the absolute path to the original file
        original_abspath = self.current_task.get_source_abspath()

        # Ensure file exists
        if not os.path.exists(original_abspath):
            self._log("File no longer exists '{}'. Was it removed?".format(original_abspath), level='warning')
            self.__write_failure_to_worker_log()
            return False

        # Set the remote worker address
        address = self.installation_info.get('address')

        lock_key = None

        # Fetch the library name and path this task is for
        library_id = self.current_task.get_task_library_id()
        try:
            library = Library(library_id)
        except Exception as e:
            self._log("Unable to fetch library config for ID {}".format(library_id), level='exception')
            self.__write_failure_to_worker_log()
            return False
        library_name = library.get_name()
        library_path = library.get_path()

        # Check if we can create the remote task with just a relative path
        #   only create checksum and send file if the remote library path cannot accept relative paths or
        #   it is configured for only receiving remote files
        send_file = False
        library_config = self.links.get_the_remote_library_config_by_name(self.installation_info, library_name)

        # Check if remote library is configured only for receiving remote files
        if library_config.get('enable_remote_only'):
            send_file = True

        # First attempt to create a task with an abspath on the remote installation
        remote_task_id = None
        if not send_file:
            remote_library_id = library_config.get('id')

            # Remove library path from file abspath to create a relative path
            original_relpath = os.path.relpath(original_abspath, library_path)
            # Join remote library path to the relative path to form a remote library abspath to the file
            remote_original_abspath = os.path.join(library_config.get('path'), original_relpath)
            # Post the task creation. This will error if the file does not exist
            info = self.links.new_pending_task_create_on_remote_installation(self.installation_info,
                                                                             remote_original_abspath,
                                                                             remote_library_id)
            if not info:
                self._log("Unable to create remote pending task for path '{}'. Fallback to sending file.".format(
                    remote_original_abspath), level='debug')
                send_file = True
            elif 'path does not exist' in info.get('error', '').lower():
                self._log("Unable to find file in remote library's path '{}'. Fallback to sending file.".format(
                    remote_original_abspath), level='debug')
                send_file = True
            elif 'task already exists' in info.get('error', '').lower():
                self._log("A remote task already exists with the path '{}'. Fallback to sending file.".format(
                    remote_original_abspath), level='error')
                self.__write_failure_to_worker_log()
                return False

            # Set the remote task ID
            remote_task_id = info.get('id')

        if send_file:
            initial_checksum = None
            if self.installation_info.get('enable_checksum_validation', False):
                # Get source file checksum
                initial_checksum = common.get_file_checksum(original_abspath)
            initial_file_size = os.path.getsize(original_abspath)

            # Loop until we are able to upload the file to the remote installation
            info = {}
            while not self.redundant_flag.is_set():
                # For files smaller than 100MB, just transfer them in parallel
                # Smaller files add a lot of time overhead with the waiting in line and it slows the whole process down
                # Larger files benefit from being transferred one at a time.
                if initial_file_size > 100000000:
                    # Check for network transfer lock
                    lock_key = self.links.acquire_network_transfer_lock(address, transfer_limit=1, lock_type='send')
                    if not lock_key:
                        self.event.wait(1)
                        continue

                # Send a file to a remote installation.
                self._log("Uploading file to remote installation '{}'".format(original_abspath), level='debug')
                info = self.links.send_file_to_remote_installation(self.installation_info, original_abspath)
                self.links.release_network_transfer_lock(lock_key)
                if not info:
                    self._log("Failed to upload the file '{}'".format(original_abspath), level='error')
                    self.__write_failure_to_worker_log()
                    return False
                break

            # Set the remote task ID
            remote_task_id = info.get('id')

            # Compare uploaded file md5checksum
            if initial_checksum and info.get('checksum') != initial_checksum:
                self._log("The uploaded file did not return a correct checksum '{}'".format(original_abspath), level='error')
                # Send request to terminate the remote worker then return
                self.links.remove_task_from_remote_installation(self.installation_info, remote_task_id)
                self.__write_failure_to_worker_log()
                return False

        # Ensure at this point we have set the remote_task_id
        if remote_task_id is None:
            self._log("Failed to create remote task. Var remote_task_id is still None", level='error')
            self.__write_failure_to_worker_log()
            return False

        # Set the library of the remote task using the library's name
        while not self.redundant_flag.is_set():
            result = self.links.set_the_remote_task_library(self.installation_info, remote_task_id, library_name)
            if result is None:
                # Unable to reach remote installation
                self.event.wait(2)
                continue
            if not result.get('success'):
                self._log(
                    "Failed to match a remote library named '{}'. Remote installation will use the default library".format(
                        library_name), level='warning')
                # Just log the warning for this. If no matching library name is found it will remain set as the default library
                break
            if result.get('success'):
                break

        # Start the remote task
        while not self.redundant_flag.is_set():
            result = self.links.start_the_remote_task_by_id(self.installation_info, remote_task_id)
            if not result:
                # Unable to reach remote installation
                self.event.wait(2)
                continue
            if not result.get('success'):
                self._log("Failed to set initial remote pending task to status '{}'".format(original_abspath), level='error')
                # Send request to terminate the remote worker then return
                self.links.remove_task_from_remote_installation(self.installation_info, remote_task_id)
                self.__write_failure_to_worker_log()
                return False
            if result.get('success'):
                break

        # Loop while redundant_flag not set (while true because of below)
        worker_id = None
        task_status = ''
        last_status_fetch = 0
        polling_delay = 5
        while task_status != 'complete':
            self.event.wait(1)
            if self.redundant_flag.is_set():
                # Send request to terminate the remote worker then exit
                if worker_id:
                    self.links.terminate_remote_worker(self.installation_info, worker_id)
                break

            # Only fetch the status every 5 seconds
            time_now = time.time()
            if last_status_fetch > (time_now - polling_delay):
                continue

            # Fetch task status
            all_task_states = self.links.get_remote_pending_task_state(self.installation_info, remote_task_id)
            task_status = ''
            polling_delay = 5
            if all_task_states:
                for ts in all_task_states.get('results', []):
                    if str(ts.get('id')) == str(remote_task_id):
                        # Task is complete. Exit loop but do not set redundant flag on link manager
                        task_status = ts.get('status')
                        break
                if not all_task_states.get('results', []):
                    # Remote task list is empty
                    task_status = 'removed'
                elif all_task_states.get('results') and task_status == '':
                    # Remote task list did not contain this task
                    task_status = 'removed'

            # If the task status is 'complete', break the loop here and move onto the result retrieval
            # If all_task_states returned no results (we are unable to connect to the remote installation)
            # If all_task_states did return results but our task_status was found, the remote installation has removed our task
            # If the task status is not 'in_progress', loop here and wait for task to be picked up by a worker
            if task_status == 'complete':
                break
            elif not all_task_states:
                polling_delay = 10
                last_status_fetch = time_now
                continue
            elif task_status == 'removed':
                self._log("Task has been removed by remote installation '{}'".format(original_abspath), level='error')
                self.__write_failure_to_worker_log()
                return False
            elif task_status != 'in_progress':
                # Mark this as the last time run
                last_status_fetch = time_now
                polling_delay = 10
                continue

            # Check if we know the task's worker ID already
            if not worker_id:
                # The task has been picked up by a worker, find out which one...
                workers_status = self.links.get_all_worker_status(self.installation_info)
                if not workers_status:
                    # The request failed for some reason... Perhaps we lost contact with the remote installation
                    # Mark this as the last time run
                    last_status_fetch = time_now
                    continue
                for worker in workers_status:
                    if str(worker.get('current_task')) == str(remote_task_id):
                        worker_id = worker.get('id')

            # Fetch worker progress
            worker_status = self.links.get_single_worker_status(self.installation_info, worker_id)
            if not worker_status:
                # Mark this as the last time run
                last_status_fetch = time_now
                continue

            # Update status
            self.paused = worker_status.get('paused')
            self.worker_log = worker_status.get('worker_log_tail')
            self.worker_runners_info = worker_status.get('runners_info')
            self.worker_subprocess_percent = worker_status.get('subprocess', {}).get('percent')
            self.worker_subprocess_elapsed = worker_status.get('subprocess', {}).get('elapsed')

            # Mark this as the last time run
            last_status_fetch = time_now

        # If the previous loop was broken because this tread needs to terminate, return False here (did not complete)
        if self.redundant_flag.is_set():
            self.worker_log += ["\n\nREMOTE LINK MANAGER TERMINATED!"]
            self.current_task.save_command_log(self.worker_log)
            return False

        self._log("Remote task completed '{}'".format(original_abspath), level='info')

        # Create local cache path to download results
        task_cache_path = self.current_task.get_cache_path()
        # Ensure the cache directory exists
        cache_directory = os.path.dirname(os.path.abspath(task_cache_path))
        if not os.path.exists(cache_directory):
            os.makedirs(cache_directory)

        # Fetch remote task result data
        data = self.links.fetch_remote_task_data(self.installation_info, remote_task_id,
                                                 os.path.join(cache_directory, 'remote_data.json'))

        if not data:
            self._log(
                "Failed to retrieve remote task data for '{}'. NOTE: The cached files have not been removed from the remote host.".format(
                    original_abspath), level='error')
            self.__write_failure_to_worker_log()
            return False
        self.worker_log = [data.get('log')]

        # Save the completed command log
        self.current_task.save_command_log(self.worker_log)

        # Fetch remote task file
        if data.get('task_success'):
            task_label = data.get('task_label')
            self._log(
                "Remote task #{} was successful, proceeding to download the completed file '{}'".format(remote_task_id,
                                                                                                        task_label),
                level='debug')
            # Set the new file out as the extension may have changed
            split_file_name = os.path.splitext(data.get('abspath'))
            file_extension = split_file_name[1].lstrip('.')
            self.current_task.set_cache_path(cache_directory, file_extension)
            # Read the updated cache path
            task_cache_path = self.current_task.get_cache_path()

            # Loop until we are able to upload the file to the remote installation
            while not self.redundant_flag.is_set():
                # Check for network transfer lock
                lock_key = self.links.acquire_network_transfer_lock(address, transfer_limit=2, lock_type='receive')
                if not lock_key:
                    self.event.wait(1)
                    continue
                # Download the file
                self._log("Downloading file from remote installation '{}'".format(task_label), level='debug')
                success = self.links.fetch_remote_task_completed_file(self.installation_info, remote_task_id, task_cache_path)
                self.links.release_network_transfer_lock(lock_key)
                if not success:
                    self._log("Failed to download file '{}'".format(os.path.basename(data.get('abspath'))), level='error')
                    # Send request to terminate the remote worker then return
                    self.links.remove_task_from_remote_installation(self.installation_info, remote_task_id)
                    self.__write_failure_to_worker_log()
                    return False
                break

            # If the previous loop was broken because this tread needs to terminate, return False here (did not complete)
            if self.redundant_flag.is_set():
                self.worker_log += ["\n\nREMOTE LINK MANAGER TERMINATED!"]
                self.current_task.save_command_log(self.worker_log)
                return False

            # Match checksum from task result data with downloaded file
            if self.installation_info.get('enable_checksum_validation', False):
                downloaded_checksum = common.get_file_checksum(task_cache_path)
                if downloaded_checksum != data.get('checksum'):
                    self._log("The downloaded file did not produce a correct checksum '{}'".format(task_cache_path),
                              level='error')
                    # Send request to terminate the remote worker then return
                    self.links.remove_task_from_remote_installation(self.installation_info, remote_task_id)
                    self.__write_failure_to_worker_log()
                    return False

            # Send request to terminate the remote worker then return
            self.links.remove_task_from_remote_installation(self.installation_info, remote_task_id)

            return True

        self.__write_failure_to_worker_log()
        return False
